from enum import Enum


class APITerms(Enum):
    """
    Defines common endpoint terms used in the API
    """
    API = 'api'
    COLLECTIONS = 'collections'
    COMMANDS = 'commands'
    COMPONENTS = 'components'
    CONFORMANCE = 'conformance'
    CONTROL_STREAMS = 'controlstreams'
    DATASTREAMS = 'datastreams'
    DEPLOYMENTS = 'deployments'
    EVENTS = 'events'
    FOIS = 'featuresOfInterest'
    HISTORY = 'history'
    ITEMS = 'items'
    OBSERVATIONS = 'observations'
    PROCEDURES = 'procedures'
    PROPERTIES = 'properties'
    SAMPLING_FEATURES = 'samplingFeatures'
    SCHEMA = 'schema'
    STATUS = 'status'
    SYSTEMS = 'systems'
    SYSTEM_EVENTS = 'systemEvents'
    TASKING = 'controls'
    UNDEFINED = ''


class SystemTypes(Enum):
    """
    Defines the system types
    """
    FEATURE = "Feature"


class ObservationFormat(Enum):
    """
    Defines common observation formats
    """
    JSON = "application/om+json"
    XML = "application/om+xml"
    SWE_XML = "application/swe+xml"
    SWE_JSON = "application/swe+json"
    SWE_CSV = "application/swe+csv"
    SWE_BINARY = "application/swe+binary"
    SWE_TEXT = "application/swe+text"


class DatastreamResultTypes(Enum):
    """
    Defines the datastream result types
    """
    MEASURE = "measure"
    VECTOR = "vector"
    RECORD = "record"
    COVERAGE_1D = "coverage1D"
    COVERAGE_2D = "coverage2D"
    COVERAGE_3D = "coverage3D"


class GeometryTypes(Enum):
    """
    Defines the geometry types
    """
    POINT = "Point"
    LINESTRING = "LineString"
    POLYGON = "Polygon"
    MULTI_POINT = "MultiPoint"
    MULTI_LINESTRING = "MultiLineString"
    MULTI_POLYGON = "MultiPolygon"


class APIResourceTypes(Enum):
    """
    Defines the resource types
    """
    ROOT = ""
    COLLECTION = "Collection"
    COMMAND = "Command"
    COMPONENT = "Component"
    CONTROL_CHANNEL = "ControlChannel"
    DATASTREAM = "Datastream"
    DEPLOYMENT = "Deployment"
    OBSERVATION = "Observation"
    PROCEDURE = "Procedure"
    PROPERTY = "Property"
    SAMPLING_FEATURE = "SamplingFeature"
    SYSTEM = "System"
    SYSTEM_EVENT = "SystemEvent"
    SYSTEM_HISTORY = "SystemHistory"
    STATUS = "Status"
    SCHEMA = "Schema"


class ContentTypes(Enum):
    """
    Defines the encoding formats
    """
    JSON = "application/json"
    XML = "application/xml"
    SWE_XML = "application/swe+xml"
    SWE_JSON = "application/swe+json"
    SWE_CSV = "application/swe+csv"
    SWE_BINARY = "application/swe+binary"
    SWE_TEXT = "application/swe+text"
    GEO_JSON = "application/geo+json"
    SML_JSON = "application/sml+json"
    OM_JSON = "application/om+json"



#  =============================================================================
#  Copyright (c) 2025 Botts Innovative Research Inc.
#  Date: 2025/9/30
#  Author: Ian Patterson
#  Contact Email: ian@botts-inc.com
#  =============================================================================
from __future__ import annotations

from datetime import datetime
from typing import Union, List

from pydantic import BaseModel, Field, SerializeAsAny, field_validator, HttpUrl, ConfigDict

from .api_utils import Link, URI
from .csapi4py.constants import ObservationFormat
from .encoding import Encoding
from .geometry import Geometry
from .swe_components import AnyComponentSchema

"""
In many of the top level resource models there is a "schema" field of some description. These models are meant to ease
the burden on the end user to create those.
"""


class CommandJSON(BaseModel):
    """
    A class to represent a command in JSON format
    """
    model_config = ConfigDict(populate_by_name=True)
    control_id: str = Field(None, serialization_alias="control@id")
    issue_time: Union[str, float] = Field(datetime.now().isoformat(), serialization_alias="issueTime")
    sender: str = Field(None)
    params: Union[dict, list, int, float, str] = Field(None)


class CommandSchema(BaseModel):
    """
    Base class representation for control streams' command schemas
    """
    model_config = ConfigDict(populate_by_name=True)

    command_format: str = Field(..., alias='commandFormat')


class SWEJSONCommandSchema(CommandSchema):
    """
    SWE+JSON command schema
    """
    model_config = ConfigDict(populate_by_name=True)

    command_format: str = Field("application/swe+json", alias='commandFormat')
    encoding: SerializeAsAny[Encoding] = Field(...)
    record_schema: SerializeAsAny[AnyComponentSchema] = Field(..., serialization_alias='recordSchema')


class JSONCommandSchema(CommandSchema):
    """
    JSON command schema
    """
    model_config = ConfigDict(populate_by_name=True)

    command_format: str = Field("application/json", alias='commandFormat')
    params_schema: SerializeAsAny[AnyComponentSchema] = Field(..., alias='parametersSchema')
    result_schema: SerializeAsAny[AnyComponentSchema] = Field(None, alias='resultSchema')
    feasibility_schema: SerializeAsAny[AnyComponentSchema] = Field(None, alias='feasibilityResultSchema')


class DatastreamRecordSchema(BaseModel):
    """
    A class to represent the schema of a datastream
    """
    model_config = ConfigDict(populate_by_name=True)

    obs_format: str = Field(..., alias='obsFormat')


class SWEDatastreamRecordSchema(DatastreamRecordSchema):
    model_config = ConfigDict(populate_by_name=True)
    encoding: SerializeAsAny[Encoding] = Field(...)
    record_schema: SerializeAsAny[AnyComponentSchema] = Field(..., serialization_alias='recordSchema')

    @field_validator('obs_format')
    @classmethod
    def check_check_obs_format(cls, v):
        if v not in [ObservationFormat.SWE_JSON.value, ObservationFormat.SWE_CSV.value,
                     ObservationFormat.SWE_TEXT.value, ObservationFormat.SWE_BINARY.value]:
            raise ValueError('obsFormat must be on of the SWE formats')
        return v


class ObservationOMJSONInline(BaseModel):
    """
    A class to represent an observation in OM-JSON format
    """
    model_config = ConfigDict(populate_by_name=True)
    datastream_id: str = Field(None, serialization_alias="datastream@id")
    foi_id: str = Field(None, serialization_alias="foi@id")
    phenomenon_time: str = Field(None, serialization_alias="phenomenonTime")
    result_time: str = Field(datetime.now().isoformat(), serialization_alias="resultTime")
    parameters: dict = Field(None)
    result: Union[int, float, str, dict, list] = Field(...)
    result_links: List[Link] = Field(None, serialization_alias="result@links")


class SystemEventOMJSON(BaseModel):
    """
    A class to represent the schema of a system event
    """
    model_config = ConfigDict(populate_by_name=True)
    label: str = Field(...)
    description: str = Field(None)
    definition: HttpUrl = Field(...)
    identifiers: list = Field(None)
    classifiers: list = Field(None)
    contacts: list = Field(None)
    documentation: list = Field(None)
    time: str = Field(...)
    properties: list = Field(None)
    configuration: dict = Field(None)
    links: list[Link] = Field(None)


class SystemHistoryGeoJSON(BaseModel):
    """
    A class to represent the schema of a system history
    """
    model_config = ConfigDict(populate_by_name=True)
    type: str = Field(...)
    id: str = Field(None)
    properties: SystemHistoryProperties = Field(...)
    geometry: Geometry = Field(None)
    bbox: list = Field(None)
    links: list[Link] = Field(None)


class SystemHistoryProperties(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    feature_type: str = Field(...)
    uid: URI = Field(...)
    name: str = Field(...)
    description: str = Field(None)
    asset_type: str = Field(None)
    valid_time: list = Field(None)
    parent_system_link: str = Field(None, serialization_alias='parentSystem@link')
    procedure_link: str = Field(None, serialization_alias='procedure@link')



#  =============================================================================
#  Copyright (c) 2025 Botts Innovative Research Inc.
#  Date: 2025/10/22
#  Author: Ian Patterson
#  Contact Email: ian@botts-inc.com
#  =============================================================================
from src.oshconnect.resource_datamodels import ControlStreamResource


def test_control_stream_resource():
    res_str = {'id': '0228vl6tn15g', 'name': 'Puppy Pi Control', 'description': 'Puppy pi control', 'system@id': '029tjlvogsng', 'system@link': {'href': 'http://192.168.8.136:8080/sensorhub/api/systems/029tjlvogsng?f=json', 'uid': 'urn:puppypi:001', 'type': 'application/geo+json'}, 'inputName': 'puppypicontrol', 'validTime': ['2025-10-21T19:04:56.505817Z', 'now'], 'issueTime': ['2025-10-22T17:12:58.51182Z', '2025-10-22T17:12:58.51182Z'], 'controlledProperties': [{'definition': 'http://sensorml.com/ont/swe/property/triggercontrol', 'label': 'Forward', 'description': 'Moves the puppy pi forward when true'}], 'formats': ['application/json', 'application/swe+json', 'application/swe+csv', 'application/swe+xml', 'application/swe+binary']}
    # res_dict = json.loads(res_str)
    csr = ControlStreamResource.model_validate(res_str)

    assert isinstance(csr, ControlStreamResource)



from enum import Enum

import requests
# import websockets
from pydantic import BaseModel, Field

from .constants import APITerms


class Endpoint(BaseModel):
    api_root: str = APITerms.API.value
    base_resource: APITerms = Field(None)
    resource_id: str = Field(None)
    sub_component: APITerms = Field(None)
    secondary_resource_id: str = Field(None)

    def create_endpoint(self):
        # TODO: Handle insertion of  "/" in the right places
        # Create endpoints bases of api spec
        base_res_id = '' if self.base_resource is None else f'/{self.base_resource}'
        res_id = '' if self.resource_id is None else f'/{self.resource_id}'
        sub_comp = '' if self.sub_component is None else f'/{self.sub_component}'
        secondary_res_id = '' if self.secondary_resource_id is None else f'/{self.secondary_resource_id}'

        return f'{self.api_root}{base_res_id}{res_id}{sub_comp}{secondary_res_id}'


class SystemQueryParams(Enum):
    Keywords = 'q'
    """
    A comma-separated list of keywords to search for in the system name, description, and definition.
    """
    BBOX = 'bbox'
    """
    BBOX to fileter resources based on their location
    """
    LOCATION = 'location'
    """
    WKT geometry to filter resources based on their location or geometry
    """
    VALID_TIME = 'validTime'
    """
    ISO 8601 time interval to filter resources based on their valid time. When omitted, the implicit time is "now"
    except for "history" collection where no filtering is applied.
    """
    PARENT = 'parent'
    """
    Comma-separated list of parent system IDs or "*" to included nested resources at any level
    """
    SELECT = 'select'
    """
    Comma-separated list of properties to include or exclude from results (use "!" prefix to exclude)
    """
    FORMAT = 'format'
    """
    Mime type of the response format.
    """
    LIMIT = 'limit'
    """
    Maximum number of resources to return per page (max 1000)
    """
    OFFSET = 'offset'
    """
    Token specifying the page to return (usually the token provided in the previous call)
    """

    @classmethod
    def list(cls):
        return list(map(lambda c: c.value, cls))


def handle_request(url, params=None, content_json=None, method='get', response_handler=None):
    """
    Handles a request to the API
    :param url: The URL to make the request to
    :param params: The parameters to send with the request
    :param content_json: The JSON to send with the request
    :param method: The method to use for the request
    :param response_handler:
    :return: The response from the API
    """

    r = None

    if method == 'get':
        r = requests.get(url, params=params)
    elif method == 'post':
        r = requests.post(url, params=params, json=content_json, headers={'Content-Type': 'application/json'})
    elif method == 'put':
        r = requests.put(url, params=params, json=content_json)
    elif method == 'delete':
        r = requests.delete(url, params=params)
    else:
        raise ValueError(f'Invalid method: {method}')

    if response_handler is not None:
        return response_handler(r)
    else:
        return r


# async def handle_ws(url, params=None, json_data=None, method='get', response_handler=None):
#     """
#     Handles a request to the API. Functionality is limited to receiving observations for now, but will be improved in
#     future versions.
#     :param url: The URL to make the request to
#     :param params: The parameters to send with the request
#     :param json_data: The JSON to send with the request
#     :param method: The method to use for the request
#     :param response_handler: callback function to handle the response msg
#     :return: The response from the API
#     """
#
#     async with websockets.connect(url) as ws:
#         while True:
#             msg = await ws.recv()
#             response_handler(msg)



from pydantic import BaseModel


def serialize_model_list(model_list: list[BaseModel]):
    """
    Serializes a list of pydantic models
    :param model_list:
    :return: a valid json string
    """
    return '[' + ','.join([model.model_dump_json(exclude_none=True, by_alias=True) for model in model_list]) + ']'



#  =============================================================================
#  Copyright (c) 2025 Botts Innovative Research Inc.
#  Date: 2025/9/30
#  Author: Ian Patterson
#  Contact Email: ian@botts-inc.com
#  =============================================================================

from __future__ import annotations

from abc import ABC
from dataclasses import dataclass

from pydantic import BaseModel, Field

from .con_sys_api import ConnectedSystemAPIRequest
from .constants import APIResourceTypes, ContentTypes, APITerms


# TODO: rework to make the first resource in the endpoint the primary key for URL construction, currently, the implementation is a bit on the confusing side with what is being generated and why.

def determine_parent_type(res_type: APIResourceTypes):
    match res_type:
        case APIResourceTypes.SYSTEM:
            return APIResourceTypes.SYSTEM
        case APIResourceTypes.COLLECTION:
            return None
        case APIResourceTypes.CONTROL_CHANNEL:
            return APIResourceTypes.SYSTEM
        case APIResourceTypes.COMMAND:
            return APIResourceTypes.CONTROL_CHANNEL
        case APIResourceTypes.DATASTREAM:
            return APIResourceTypes.SYSTEM
        case APIResourceTypes.OBSERVATION:
            return APIResourceTypes.DATASTREAM
        case APIResourceTypes.SYSTEM_EVENT:
            return APIResourceTypes.SYSTEM
        case APIResourceTypes.SAMPLING_FEATURE:
            return APIResourceTypes.SYSTEM
        case APIResourceTypes.PROCEDURE:
            return None
        case APIResourceTypes.PROPERTY:
            return None
        case APIResourceTypes.SYSTEM_HISTORY:
            return None
        case APIResourceTypes.DEPLOYMENT:
            return None
        case _:
            return None


def resource_type_to_endpoint(res_type: APIResourceTypes, parent_type: APIResourceTypes = None):
    if parent_type is APIResourceTypes.COLLECTION:
        return APITerms.ITEMS.value

    match res_type:
        case APIResourceTypes.SYSTEM:
            return APITerms.SYSTEMS.value
        case APIResourceTypes.COLLECTION:
            return APITerms.COLLECTIONS.value
        case APIResourceTypes.CONTROL_CHANNEL:
            return APITerms.CONTROL_STREAMS.value
        case APIResourceTypes.COMMAND:
            return APITerms.COMMANDS.value
        case APIResourceTypes.DATASTREAM:
            return APITerms.DATASTREAMS.value
        case APIResourceTypes.OBSERVATION:
            return APITerms.OBSERVATIONS.value
        case APIResourceTypes.SYSTEM_EVENT:
            return APITerms.SYSTEM_EVENTS.value
        case APIResourceTypes.SAMPLING_FEATURE:
            return APITerms.SAMPLING_FEATURES.value
        case APIResourceTypes.PROCEDURE:
            return APITerms.PROCEDURES.value
        case APIResourceTypes.PROPERTY:
            return APITerms.PROPERTIES.value
        case APIResourceTypes.SYSTEM_HISTORY:
            return APITerms.HISTORY.value
        case APIResourceTypes.DEPLOYMENT:
            return APITerms.DEPLOYMENTS.value
        case APIResourceTypes.STATUS:
            return APITerms.STATUS.value
        case APIResourceTypes.SCHEMA:
            return APITerms.SCHEMA.value
        case _:
            raise ValueError('Invalid resource type')


@dataclass
class APIHelper(ABC):
    server_url: str = None
    port: int = None
    protocol: str = "https"
    server_root: str = "sensorhub"
    api_root: str = "api"
    username: str = None
    password: str = None
    user_auth: bool = False

    def create_resource(self, res_type: APIResourceTypes, json_data: any, parent_res_id: str = None,
                        from_collection: bool = False, url_endpoint: str = None, req_headers: dict = None):
        """
        Creates a resource of the given type with the given data, will attempt to create a sub-resource if parent_res_id
        is provided.
        :param req_headers:
        :param res_type:
        :param json_data:
        :param parent_res_id:
        :param from_collection:
        :param url_endpoint: If given, will override the default URL construction. Should contain the endpoint past the API root.
        :return:
        """

        if url_endpoint is None:
            url = self.resource_url_resolver(res_type, None, parent_res_id, from_collection)
        else:
            url = f'{self.server_url}/{self.api_root}/{url_endpoint}'
        api_request = ConnectedSystemAPIRequest(url=url, request_method='POST', auth=self.get_helper_auth(),
                                                body=json_data, headers=req_headers)
        return api_request.make_request()

    def retrieve_resource(self, res_type: APIResourceTypes, res_id: str = None, parent_res_id: str = None,
                          from_collection: bool = False,
                          collection_id: str = None, url_endpoint: str = None, req_headers: dict = None):
        """
        Retrieves a resource or list of resources if no res_id is provided, will attempt to retrieve a sub-resource if
        parent_res_id is provided.
        :param req_headers:
        :param res_type:
        :param res_id:
        :param parent_res_id:
        :param from_collection:
        :param collection_id:
        :param url_endpoint: If given, will override the default URL construction. Should contain the endpoint past the API root.
        :return:
        """
        if url_endpoint is None:
            url = self.resource_url_resolver(res_type, res_id, parent_res_id, from_collection)
        else:
            url = f'{self.server_url}/{self.api_root}/{url_endpoint}'
        api_request = ConnectedSystemAPIRequest(url=url, request_method='GET', auth=self.get_helper_auth(),
                                                headers=req_headers)
        return api_request.make_request()

    def get_resource(self, resource_type: APIResourceTypes, resource_id: str = None,
                     subresource_type: APIResourceTypes = None,
                     req_headers: dict = None):

        """
        Helper to get resources by type, specifically by id, and optionally a sub-resource collection of a specified resource.
        :param resource_type:
        :param resource_id:
        :param subresource_type:
        :param req_headers:
        :return:
        """
        if req_headers is None:
            req_headers = {}
        base_api_url = self.get_api_root_url()
        resource_type_str = resource_type_to_endpoint(resource_type)
        res_id_str = f'/{resource_id}' if resource_id else ""
        sub_res_type_str = f'/{resource_type_to_endpoint(subresource_type)}' if subresource_type else ""
        complete_url = f'{base_api_url}/{resource_type_str}{res_id_str}{sub_res_type_str}'
        api_request = ConnectedSystemAPIRequest(url=complete_url, request_method='GET', auth=self.get_helper_auth(),
                                                headers=req_headers)
        return api_request.make_request()

    def update_resource(self, res_type: APIResourceTypes, res_id: str, json_data: any, parent_res_id: str = None,
                        from_collection: bool = False, url_endpoint: str = None, req_headers: dict = None):
        """
        Updates a resource of the given type by its id, if necessary, will attempt to update a sub-resource if
        parent_res_id is provided.
        :param req_headers:
        :param res_type:
        :param res_id:
        :param json_data:
        :param parent_res_id:
        :param from_collection:
        :param url_endpoint: If given, will override the default URL construction. Should contain the endpoint past the API root.
        :return:
        """
        if url_endpoint is None:
            url = self.resource_url_resolver(res_type, None, parent_res_id, from_collection)
        else:
            url = f'{self.server_url}/{self.api_root}/{url_endpoint}'
        api_request = ConnectedSystemAPIRequest(url=url, request_method='PUT', auth=self.get_helper_auth(),
                                                body=json_data, headers=req_headers)
        return api_request.make_request()

    def delete_resource(self, res_type: APIResourceTypes, res_id: str, parent_res_id: str = None,
                        from_collection: bool = False, url_endpoint: str = None, req_headers: dict = None):
        """
        Deletes a resource of the given type by its id, if necessary, will attempt to delete a sub-resource if
        parent_res_id is provided.
        :param req_headers:
        :param res_type:
        :param res_id:
        :param parent_res_id:
        :param from_collection:
        :param url_endpoint: If given, will override the default URL construction. Should contain the endpoint past the API root.
        :return:
        """
        if url_endpoint is None:
            url = self.resource_url_resolver(res_type, None, parent_res_id, from_collection)
        else:
            url = f'{self.server_url}/{self.api_root}/{url_endpoint}'
        api_request = ConnectedSystemAPIRequest(url=url, request_method='DELETE', auth=self.get_helper_auth(),
                                                headers=req_headers)
        return api_request.make_request()

    # Helpers
    def resource_url_resolver(self, subresource_type: APIResourceTypes, subresource_id: str = None,
                              resource_id: str = None,
                              from_collection: bool = False):
        """
        Helper to generate a URL endpoint for a given resource type and id by matching the resource type to an
        appropriate parent endpoint and inserting the resource ids as necessary.
        :param subresource_type:
        :param subresource_id:
        :param resource_id:
        :param from_collection:
        :return:
        """
        if subresource_type is None:
            raise ValueError('Resource type must contain a valid APIResourceType')
        if subresource_type is APIResourceTypes.COLLECTION and from_collection:
            raise ValueError('Collections are not sub-resources of other collections')

        parent_type = None
        if resource_id and not from_collection:
            parent_type = determine_parent_type(subresource_type)
        elif resource_id and from_collection:
            parent_type = APIResourceTypes.COLLECTION

        return self.construct_url(parent_type, subresource_id, subresource_type, resource_id)

    def construct_url(self, resource_type: APIResourceTypes, subresource_id, subresource_type, resource_id,
                      for_socket: bool = False):
        """
        Constructs an API endpoint url from the given parameters
        :param resource_type:
        :param subresource_id:
        :param subresource_type:
        :param resource_id:
        :param for_socket: If true, will construct a WebSocket URL (ws:// or wss://) instead of HTTP/HTTPS.
        :return:
        """
        # TODO: Test for less common cases to ensure that the URL is being constructed correctly
        base_url = self.get_api_root_url(socket=for_socket)

        resource_endpoint = resource_type_to_endpoint(subresource_type, resource_type)
        url = f'{base_url}/{resource_endpoint}'

        if resource_type:
            parent_endpoint = resource_type_to_endpoint(resource_type)
            url = f'{base_url}/{parent_endpoint}/{resource_id}/{resource_endpoint}'

        if subresource_id:
            url = f'{url}/{subresource_id}'

        return url

    def get_helper_auth(self):
        if self.user_auth:
            return self.username, self.password
        return None

    def get_base_url(self, socket: bool = False):
        if socket:
            protocol = 'ws' if self.protocol == 'http' else 'wss'
            return f'{protocol}://{self.server_url}{f":{self.port}" if self.port else ""}'
        return f'{self.protocol}://{self.server_url}{f":{self.port}" if self.port else ""}'

    def get_api_root_url(self, socket: bool = False):
        """
        Returns the full API root URL including protocol, server address, port (if applicable), and API root path.
        :param socket: If true, will return a WebSocket URL (ws:// or wss://) instead of HTTP/HTTPS.
        :return:
        """
        return f'{self.get_base_url(socket=socket)}/{self.server_root}/{self.api_root}'

    def set_protocol(self, protocol: str):
        if protocol not in ['http', 'https', 'ws', 'wss']:
            raise ValueError('Protocol must be either "http" or "https"')
        self.protocol = protocol

    # TODO: add validity checking for resource type combinations
    def get_mqtt_topic(self, resource_type, subresource_type, resource_id: str, subresource_id: str = None):
        """
        Returns the MQTT topic for the resource type, does not check for validity of the resource type combination
        :param resource_type : The API resource type of the resource that comes first in the URL, cannot be None
        :param subresource_type: The API resource type of the sub-resource that comes second in the URL, optional if there
        is no sub-resource.
        :param resource_id: The ID of the primary resource, can be none if the request is being made for all resources of
        the given type.
        :param subresource_id: The ID of the sub-resource, can be none if the request is being made for all sub-resources of
        the given type.
        :return:
        """
        subresource_endpoint = f'/{resource_type_to_endpoint(subresource_type)}'
        resource_endpoint = "" if resource_type is None else f'/{resource_type_to_endpoint(resource_type)}'
        resource_ident = "" if resource_id is None else f'/{resource_id}'
        subresource_ident = "" if subresource_id is None else f'/{subresource_id}'
        topic_locator = f'/{self.api_root}{resource_endpoint}{resource_ident}{subresource_endpoint}{subresource_ident}'
        print(f'MQTT Topic: {topic_locator}')

        return topic_locator


@dataclass(kw_only=True)
class ResponseParserHelper:
    default_object_reps: DefaultObjectRepresentations


class DefaultObjectRepresentations(BaseModel):
    """
    Intended to be used as a way to determine which formats should be used when serializing and deserializing objects.
    Should work in tandem with planned Serializer/Deserializer classes.
    """
    # Part 1
    collections: str = Field(ContentTypes.JSON.value)
    deployments: str = Field(ContentTypes.GEO_JSON.value)
    procedures: str = Field(ContentTypes.GEO_JSON.value)
    properties: str = Field(ContentTypes.SML_JSON.value)
    sampling_features: str = Field(ContentTypes.GEO_JSON.value)
    systems: str = Field(ContentTypes.GEO_JSON.value)
    # Part 2
    datastreams: str = Field(ContentTypes.JSON.value)
    observations: str = Field(ContentTypes.JSON.value)
    control_channels: str = Field(ContentTypes.JSON.value)
    commands: str = Field(ContentTypes.JSON.value)
    system_events: str = Field(ContentTypes.OM_JSON.value)
    system_history: str = Field(ContentTypes.GEO_JSON.value)
    # TODO: validate schemas for each resource to amke sure they are allowed per the spec



import paho.mqtt.client as mqtt


class MQTTCommClient:
    def __init__(self, url, port=1883, username=None, password=None, path='mqtt', client_id_suffix="", transport='tcp'):
        """
    Wraps a paho mqtt client to provide a simple interface for interacting with the mqtt server that is customized
    for this library.

    :param url: url of the mqtt server
    :param port: port the mqtt server is communicating over, default is 1883 or whichever port the main node is
    using if in websocket mode
    :param username: used if node is requiring authentication to access this service
    :param password: used if node is requiring authentication to access this service
    :param path: used for setting the path when using websockets (usually sensorhub/mqtt by default)
    """
        self.__url = url
        self.__port = port
        self.__path = path
        self.__client_id = f'oscapy_mqtt-{client_id_suffix}'
        self.__transport = transport

        self.__client = mqtt.Client(mqtt.CallbackAPIVersion.VERSION2, client_id=self.__client_id)

        if self.__transport == 'websockets':
            self.__client.ws_set_options(path=self.__path)

        if username is not None and password is not None:
            self.__client.username_pw_set(username, password)
            self.__client.tls_set(tls_version=mqtt.ssl.PROTOCOL_TLSv1_2)

        self.__client.on_connect = self.on_connect
        self.__client.on_subscribe = self.on_subscribe
        self.__client.on_message = self.on_message
        self.__client.on_publish = self.on_publish
        self.__client.on_log = self.on_log
        self.__client.on_disconnect = self.on_disconnect

        self.__is_connected = False

    @staticmethod
    def on_connect(client, userdata, flags, rc, properties):
        print(f'Connected with result code: {rc}')
        print(f'{properties}')

    @staticmethod
    def on_subscribe(client, userdata, mid, granted_qos, properties):
        print(f'Subscribed: {mid} {granted_qos}')

    @staticmethod
    def on_message(client, userdata, msg):
        print(f'{msg.payload.decode("utf-8")}')

    @staticmethod
    def on_publish(client, userdata, mid, info, properties):
        print(f'Published: {mid}')

    @staticmethod
    def on_log(client, userdata, level, buf):
        print(f'Log: {buf}')

    @staticmethod
    def on_disconnect(client, userdata, dc_flag, rc, properties):
        print(f'Client {client} disconnected: {dc_flag} {rc}')

    def connect(self, keepalive=60):
        # print(f'Connecting to {self.__url}:{self.__port}')
        self.__client.connect(self.__url, self.__port, keepalive=keepalive)

    def subscribe(self, topic, qos=0, msg_callback=None):
        """
        Subscribe to a topic, and optionally set a callback for when a message is received on that topic. To actually
        retrieve any information you must set a callback.

        :param topic: MQTT topic to subscribe to (example/topic)
        :param qos: quality of service, 0, 1, or 2
        :param msg_callback: callback with the form: callback(client, userdata, msg)
        :return:
        """
        self.__client.subscribe(topic, qos)
        if msg_callback is not None:
            self.__client.message_callback_add(topic, msg_callback)

    def publish(self, topic, payload=None, qos=0, retain=False):
        self.__client.publish(topic, payload, qos, retain=retain)

    def unsubscribe(self, topic):
        self.__client.unsubscribe(topic)

    def disconnect(self):
        self.__client.disconnect()

    def set_on_connect(self, on_connect):
        """
        Set the on_connect callback for the MQTT client.

        :param on_connect:
        :return:
        """
        self.__client.on_connect = on_connect

    def set_on_disconnect(self, on_disconnect):
        """
        Set the on_disconnect callback for the MQTT client.

        :param on_disconnect:
        :return:
        """
        self.__client.on_disconnect = on_disconnect

    def set_on_subscribe(self, on_subscribe):
        """
        Set the on_subscribe callback for the MQTT client.

        :param on_subscribe:
        :return:
        """
        self.__client.on_subscribe = on_subscribe

    def set_on_unsubscribe(self, on_unsubscribe):
        """
        Set the on_unsubscribe callback for the MQTT client.

        :param on_unsubscribe:
        :return:
        """
        self.__client.on_unsubscribe = on_unsubscribe

    def set_on_publish(self, on_publish):
        """
        Set the on_publish callback for the MQTT client.

        :param on_publish:
        :return:
        """
        self.__client.on_publish = on_publish

    def set_on_message(self, on_message):
        """
        Set the on_message callback for the MQTT client. It is recommended to set individual callbacks for each
        subscribed topic.

        :param on_message:
        :return:
        """
        self.__client.on_message = on_message

    def set_on_log(self, on_log):
        """
        Set the on_log callback for the MQTT client.

        :param on_log:
        :return:
        """
        self.__client.on_log = on_log

    def set_on_message_callback(self, sub, on_message_callback):
        """
        Set the on_message callback for a specific topic.
        :param sub:
        :param on_message_callback:
        :return:
        """
        self.__client.message_callback_add(sub, on_message_callback)

    def start(self):
        """
        Start the MQTT client in a separate thread. This is required for the client to be able to receive messages.

        :return:
        """
        self.__client.loop_start()

    def stop(self):
        """
        Stop the MQTT client.

        :return:
        """
        self.__client.loop_stop()

    def __toggle_is_connected(self):
        self.__is_connected = not self.__is_connected

    def is_connected(self):
        return self.__is_connected

    @staticmethod
    def publish_single(self, topic, msg):
        self.__client.single(topic, msg, 0)

    @staticmethod
    def publish_multiple(self, topic, msgs):
        self.__client.multiple(msgs, )

    def tls_set(self):
        self.__client.tls_set()



from __future__ import annotations
from pydantic import BaseModel, Field, HttpUrl, field_validator


class Link(BaseModel):
    href: HttpUrl = Field(...)
    rel: str = Field(None)
    type: str = Field(None)
    hreflang: str = Field(None)
    title: str = Field(None)
    uid: URI = Field(None)
    rt: URI = Field(None)
    interface: URI = Field(None, serialization_alias='if')


class UCUMCode(BaseModel):
    code: str = Field(...)
    label: str = Field(None)

    @field_validator('code', 'label')
    @classmethod
    def validate_string_fields(cls, v):
        if isinstance(v, str) and len(v) > 0:
            return v
        else:
            raise ValueError('code and label must be strings of length > 0')


class URI(BaseModel):
    href: HttpUrl = Field(...)
    label: str = Field(None)

    @field_validator('label')
    @classmethod
    def validate_label(cls, v):
        if isinstance(v, str) and len(v) > 0:
            return v
        else:
            raise ValueError('label must be strings of length > 0')



#  =============================================================================
#  Copyright (c) 2025 Botts Innovative Research Inc.
#  Date: 2025/9/30
#  Author: Ian Patterson
#  Contact Email: ian@botts-inc.com
#  =============================================================================
from typing import Union

import requests
from pydantic import HttpUrl

from csapi4py.con_sys_api import ConnectedSystemsRequestBuilder
from csapi4py.constants import APITerms
from csapi4py.request_wrappers import post_request


def get_landing_page(server_addr: HttpUrl, api_root: str = APITerms.API.value):
    """
    Returns the landing page of the API
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .build_url_from_base()
                   .build())
    return api_request


def get_conformance_info(server_addr: HttpUrl, api_root: str = APITerms.API.value):
    """
    Returns the conformance information of the API
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.CONFORMANCE.value)
                   .build_url_from_base()
                   .build())
    return api_request


def list_all_collections(server_addr: HttpUrl, api_root: str = APITerms.API.value):
    """
    List all collections
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.COLLECTIONS.value)
                   .build_url_from_base()
                   .build())
    return api_request


def retrieve_collection_metadata(server_addr: HttpUrl, collection_id: str, api_root: str = APITerms.API.value):
    """
    Retrieve a collection by its ID
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.COLLECTIONS.value)
                   .with_resource_id(collection_id)
                   .build_url_from_base()
                   .build())
    return api_request


def list_all_items_in_collection(server_addr: HttpUrl, collection_id: str, api_root: str = APITerms.API.value):
    """
    Lists all systems in the server at the default API endpoint
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.COLLECTIONS.value)
                   .with_resource_id(collection_id)
                   .for_sub_resource_type(APITerms.ITEMS.value)
                   .build_url_from_base()
                   .build())
    return api_request


def retrieve_collection_item_by_id(server_addr: HttpUrl, collection_id: str, item_id: str,
                                   api_root: str = APITerms.API.value):
    """
    Retrieves a system by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.COLLECTIONS.value)
                   .with_resource_id(collection_id)
                   .for_sub_resource_type(APITerms.ITEMS.value)
                   .with_resource_id(item_id)
                   .build_url_from_base()
                   .build())
    return api_request


def list_all_commands(server_addr: HttpUrl, api_root: str = APITerms.API.value, headers: dict = None):
    """
    Lists all commands
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.COMMANDS.value)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('GET')
                   .build())

    return api_request.make_request()


def list_commands_of_control_channel(server_addr: HttpUrl, control_channel_id: str, api_root: str = APITerms.API.value,
                                     headers=None):
    """
    Lists all commands of a control channel
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.CONTROL_STREAMS.value)
                   .with_resource_id(control_channel_id)
                   .for_sub_resource_type(APITerms.COMMANDS.value)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('GET')
                   .build())

    return api_request.make_request()


def send_commands_to_specific_control_stream(server_addr: HttpUrl, control_stream_id: str,
                                             request_body: Union[dict, str],
                                             api_root: str = APITerms.API.value, headers=None):
    """
    Sends a command to a control stream by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.CONTROL_STREAMS.value)
                   .with_resource_id(control_stream_id)
                   .for_sub_resource_type(APITerms.COMMANDS.value)
                   .with_request_body(request_body)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('POST')
                   .build())

    return api_request.make_request()


def retrieve_command_by_id(server_addr: HttpUrl, command_id: str, api_root: str = APITerms.API.value, headers=None):
    """
    Retrieves a command by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.COMMANDS.value)
                   .with_resource_id(command_id)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('GET')
                   .build())

    return api_request.make_request()


def update_command_description(server_addr: HttpUrl, command_id: str, request_body: Union[dict, str],
                               api_root: str = APITerms.API.value, headers=None):
    """
    Updates a command's description by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.COMMANDS.value)
                   .with_resource_id(command_id)
                   .with_request_body(request_body)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('PUT')
                   .build())

    return api_request.make_request()


def delete_command_by_id(server_addr: HttpUrl, command_id: str, api_root: str = APITerms.API.value, headers=None):
    """
    Deletes a command by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.COMMANDS.value)
                   .with_resource_id(command_id)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('DELETE')
                   .build())

    return api_request.make_request()


def list_command_status_reports(server_addr: HttpUrl, command_id: str, api_root: str = APITerms.API.value,
                                headers=None):
    """
    Lists all status reports of a command by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.COMMANDS.value)
                   .with_resource_id(command_id)
                   .for_sub_resource_type(APITerms.STATUS.value)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('GET')
                   .build())

    return api_request.make_request()


def add_command_status_reports(server_addr: HttpUrl, command_id: str, request_body: Union[dict, str],
                               api_root: str = APITerms.API.value, headers=None):
    """
    Adds a status report to a command by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.COMMANDS.value)
                   .with_resource_id(command_id)
                   .for_sub_resource_type(APITerms.STATUS.value)
                   .with_request_body(request_body)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('POST')
                   .build())

    return api_request.make_request()


def retrieve_command_status_report_by_id(server_addr: HttpUrl, command_id: str, status_report_id: str,
                                         api_root: str = APITerms.API.value, headers=None):
    """
    Retrieves a status report of a command by its id and status report id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.COMMANDS.value)
                   .with_resource_id(command_id)
                   .for_sub_resource_type(APITerms.STATUS.value)
                   .with_secondary_resource_id(status_report_id)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('GET')
                   .build())

    return api_request.make_request()


def update_command_status_report_by_id(server_addr: HttpUrl, command_id: str, status_report_id: str,
                                       request_body: Union[dict, str], api_root: str = APITerms.API.value,
                                       headers=None):
    """
    Updates a status report of a command by its id and status report id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.COMMANDS.value)
                   .with_resource_id(command_id)
                   .for_sub_resource_type(APITerms.STATUS.value)
                   .with_secondary_resource_id(status_report_id)
                   .with_request_body(request_body)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('PUT')
                   .build())

    return api_request.make_request()


def delete_command_status_report_by_id(server_addr: HttpUrl, command_id: str, status_report_id: str,
                                       api_root: str = APITerms.API.value, headers=None):
    """
    Deletes a status report of a command by its id and status report id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.COMMANDS.value)
                   .with_resource_id(command_id)
                   .for_sub_resource_type(APITerms.STATUS.value)
                   .with_secondary_resource_id(status_report_id)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('DELETE')
                   .build())

    return api_request.make_request()


def list_all_control_streams(server_addr: HttpUrl, api_root: str = APITerms.API.value, headers: dict = None):
    """
    Lists all control streams
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.CONTROL_STREAMS.value)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('GET')
                   .build())
    return api_request.make_request()


def list_control_streams_of_system(server_addr: HttpUrl, system_id: str, api_root: str = APITerms.API.value,
                                   headers=None):
    """
    Lists all control streams of a system
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.SYSTEMS.value)
                   .with_resource_id(system_id)
                   .for_sub_resource_type(APITerms.CONTROL_STREAMS.value)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('GET')
                   .build())
    return api_request.make_request()


def add_control_streams_to_system(server_addr: HttpUrl, system_id: str, request_body: Union[str, dict],
                                  api_root: str = APITerms.API.value, headers=None):
    """
    Adds a control stream to a system by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.SYSTEMS.value)
                   .with_resource_id(system_id)
                   .for_sub_resource_type(APITerms.CONTROL_STREAMS.value)
                   .with_request_body(request_body)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('POST')
                   .build())
    return api_request.make_request()


def retrieve_control_stream_description_by_id(server_addr: HttpUrl, control_stream_id: str,
                                              api_root: str = APITerms.API.value, headers: dict = None):
    """
    Retrieves a control stream by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.CONTROL_STREAMS.value)
                   .with_resource_id(control_stream_id)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('GET')
                   .build())

    return api_request.make_request()


def update_control_stream_description_by_id(server_addr: HttpUrl, control_stream_id: str,
                                            request_body: Union[str, dict],
                                            api_root: str = APITerms.API.value, headers: dict = None):
    """
    Updates a control stream by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.CONTROL_STREAMS.value)
                   .with_resource_id(control_stream_id)
                   .with_request_body(request_body)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('PUT')
                   .build())
    return api_request.make_request()


def delete_control_stream_by_id(server_addr: HttpUrl, control_stream_id: str, api_root: str = APITerms.API.value,
                                headers=None):
    """
    Deletes a control stream by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.CONTROL_STREAMS.value)
                   .with_resource_id(control_stream_id)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('DELETE')
                   .build())

    return api_request.make_request()


def retrieve_control_stream_schema_by_id(server_addr: HttpUrl, control_stream_id: str,
                                         api_root: str = APITerms.API.value, headers: dict = None):
    """
    Retrieves a control stream schema by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.CONTROL_STREAMS.value)
                   .with_resource_id(control_stream_id)
                   .for_sub_resource_type(APITerms.SCHEMA.value)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('GET')
                   .build())

    return api_request.make_request()


def update_control_stream_schema_by_id(server_addr: HttpUrl, control_stream_id: str, request_body: Union[str, dict],
                                       api_root: str = APITerms.API.value, headers: dict = None):
    """
    Updates a control stream schema by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.CONTROL_STREAMS.value)
                   .with_resource_id(control_stream_id)
                   # .for_sub_resource_type(APITerms.SCHEMA.value)
                   .with_request_body(request_body)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('PUT')
                   .build())
    return api_request.make_request()


def list_all_datastreams(server_addr: HttpUrl, api_root: str = APITerms.API.value, headers: dict = None):
    """
    Lists all datastreams
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.DATASTREAMS.value)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('GET')
                   .build())

    return api_request.make_request()


def list_all_datastreams_of_system(server_addr: HttpUrl, system_id: str, api_root: str = APITerms.API.value,
                                   headers=None):
    """
    Lists all datastreams of a system
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.SYSTEMS.value)
                   .with_resource_id(system_id)
                   .for_sub_resource_type(APITerms.DATASTREAMS.value)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('GET')
                   .build())
    return api_request.make_request()


def add_datastreams_to_system(server_addr: HttpUrl, system_id: str, request_body: Union[str, dict],
                              api_root: str = APITerms.API.value, headers=None):
    """
    Adds a datastream to a system by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.SYSTEMS.value)
                   .with_resource_id(system_id)
                   .for_sub_resource_type(APITerms.DATASTREAMS.value)
                   .with_request_body(request_body)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('POST')
                   .build())
    return api_request.make_request()


def retrieve_datastream_by_id(server_addr: HttpUrl, datastream_id: str, api_root: str = APITerms.API.value,
                              headers=None):
    """
    Retrieves a datastream by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.DATASTREAMS.value)
                   .with_resource_id(datastream_id)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('GET')
                   .build())
    return api_request.make_request()


def update_datastream_by_id(server_addr: HttpUrl, datastream_id: str, request_body: Union[str, dict],
                            api_root: str = APITerms.API.value, headers=None):
    """
    Updates a datastream by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.DATASTREAMS.value)
                   .with_resource_id(datastream_id)
                   .with_request_body(request_body)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('PUT')
                   .build())
    return api_request.make_request()


def delete_datastream_by_id(server_addr: HttpUrl, datastream_id: str, api_root: str = APITerms.API.value, headers=None):
    """
    Deletes a datastream by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.DATASTREAMS.value)
                   .with_resource_id(datastream_id)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('DELETE')
                   .build())
    return api_request.make_request()


def retrieve_datastream_schema(server_addr: HttpUrl, datastream_id: str, api_root: str = APITerms.API.value,
                               headers=None):
    """
    Retrieves a datastream schema by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.DATASTREAMS.value)
                   .with_resource_id(datastream_id)
                   .for_sub_resource_type(APITerms.SCHEMA.value)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('GET')
                   .build())
    return api_request.make_request()


def update_datastream_schema(server_addr: HttpUrl, datastream_id: str, request_body: dict,
                             api_root: str = APITerms.API.value, headers=None):
    """
    Updates a datastream schema by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.DATASTREAMS.value)
                   .with_resource_id(datastream_id)
                   .for_resource_type(APITerms.SCHEMA.value)
                   .with_request_body(request_body)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('PUT')
                   .build())
    return api_request.make_request()


def list_all_deployments(server_addr: HttpUrl, api_root: str = APITerms.API.value, headers: dict = None):
    """
    Lists all deployments in the server at the default API endpoint
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.DEPLOYMENTS.value)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('GET')
                   .build())
    return api_request.make_request()


def create_new_deployments(server_addr: HttpUrl, request_body: Union[str, dict], api_root: str = APITerms.API.value,
                           headers: dict = None):
    """
    Create a new deployment as defined by the request body
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.DEPLOYMENTS.value)
                   .with_request_body(request_body)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('POST')
                   .build())
    return api_request.make_request()


def retrieve_deployment_by_id(server_addr: HttpUrl, deployment_id: str, api_root: str = APITerms.API.value,
                              headers: dict = None):
    """
    Retrieve a deployment by its ID
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.DEPLOYMENTS.value)
                   .with_resource_id(deployment_id)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('GET')
                   .build())
    return api_request.make_request()


def update_deployment_by_id(server_addr: HttpUrl, deployment_id: str, request_body: Union[str, dict],
                            api_root: str = APITerms.API.value, headers: dict = None):
    """
    Update a deployment by its ID
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.DEPLOYMENTS.value)
                   .with_resource_id(deployment_id)
                   .with_request_body(request_body)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('PUT')
                   .build())
    return api_request.make_request()


def delete_deployment_by_id(server_addr: HttpUrl, deployment_id: str, api_root: str = APITerms.API.value,
                            headers: dict = None):
    """
    Delete a deployment by its ID
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.DEPLOYMENTS.value)
                   .with_resource_id(deployment_id)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('DELETE')
                   .build())
    return api_request


def list_deployed_systems(server_addr: HttpUrl, deployment_id, api_root: str = APITerms.API.value,
                          headers: dict = None):
    """
    Lists all deployed systems in the server at the default API endpoint
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.DEPLOYMENTS.value)
                   .with_resource_id(deployment_id)
                   .for_sub_resource_type(APITerms.SYSTEMS.value)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('GET')
                   .build())
    return api_request.make_request()


def add_systems_to_deployment(server_addr: HttpUrl, deployment_id: str, uri_list: str,
                              api_root: str = APITerms.API.value, headers: dict = None):
    """
    Lists all systems in the server at the default API endpoint
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.DEPLOYMENTS.value)
                   .with_resource_id(deployment_id)
                   .for_sub_resource_type(APITerms.SYSTEMS.value)
                   .with_request_body(uri_list)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('POST')
                   .build())
    return api_request.make_request()


def retrieve_deployed_system_by_id(server_addr: HttpUrl, deployment_id: str, system_id: str,
                                   api_root: str = APITerms.API.value, headers: dict = None):
    """
    Retrieves a system by its id
    :return:
    """

    # TODO: Add a way to have a secondary resource ID for certain endpoints
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.DEPLOYMENTS.value)
                   .with_resource_id(deployment_id)
                   .for_sub_resource_type(APITerms.SYSTEMS.value)
                   .with_secondary_resource_id(system_id)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('GET')
                   .build())
    return api_request.make_request()


def update_deployed_system_by_id(server_addr: HttpUrl, deployment_id: str, system_id: str, request_body: dict,
                                 api_root: str = APITerms.API.value, headers: dict = None):
    """
    Update a system by its ID
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.DEPLOYMENTS.value)
                   .with_resource_id(deployment_id)
                   .for_sub_resource_type(APITerms.SYSTEMS.value)
                   .with_secondary_resource_id(system_id)
                   .with_request_body(request_body)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('PUT')
                   .build())

    return api_request


def delete_deployed_system_by_id(server_addr: HttpUrl, deployment_id: str, system_id: str,
                                 api_root: str = APITerms.API.value, headers: dict = None):
    """
    Delete a system by its ID
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.DEPLOYMENTS.value)
                   .with_resource_id(deployment_id)
                   .for_sub_resource_type(APITerms.SYSTEMS.value)
                   .with_secondary_resource_id(system_id)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('DELETE')
                   .build())
    return api_request.make_request()


def list_deployments_of_specific_system(server_addr: HttpUrl, system_id: str, api_root: str = APITerms.API.value,
                                        headers: dict = None):
    """
    Lists all deployments of a specific system in the server at the default API endpoint
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.SYSTEMS.value)
                   .with_resource_id(system_id)
                   .for_sub_resource_type(APITerms.DEPLOYMENTS.value)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('GET')
                   .build())
    return api_request.make_request()


def list_all_observations(server_addr: HttpUrl, api_root: str = APITerms.API.value, headers=None):
    """
    Lists all observations
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.OBSERVATIONS.value)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('GET')
                   .build())

    return api_request.make_request()


def list_observations_from_datastream(server_addr: HttpUrl, datastream_id: str, api_root: str = APITerms.API.value,
                                      headers=None):
    """
    Lists all observations of a datastream
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.DATASTREAMS.value)
                   .with_resource_id(datastream_id)
                   .for_sub_resource_type(APITerms.OBSERVATIONS.value)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('GET')
                   .build())

    return api_request.make_request()


def add_observations_to_datastream(server_addr: HttpUrl, datastream_id: str, request_body: Union[str, dict],
                                   api_root: str = APITerms.API.value, headers=None):
    """
    Adds an observation to a datastream by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.DATASTREAMS.value)
                   .with_resource_id(datastream_id)
                   .for_sub_resource_type(APITerms.OBSERVATIONS.value)
                   .with_request_body(request_body)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('POST')
                   .build())

    return api_request.make_request()


def retrieve_observation_by_id(server_addr: HttpUrl, observation_id: str, api_root: str = APITerms.API.value,
                               headers=None):
    """
    Retrieves an observation by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.OBSERVATIONS.value)
                   .with_resource_id(observation_id)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('GET')
                   .build())

    return api_request.make_request()


def update_observation_by_id(server_addr: HttpUrl, observation_id: str, request_body: Union[str, dict],
                             api_root: str = APITerms.API.value, headers=None):
    """
    Updates an observation by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.OBSERVATIONS.value)
                   .with_resource_id(observation_id)
                   .with_request_body(request_body)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('PUT')
                   .build())

    return api_request.make_request()


def delete_observation_by_id(server_addr: HttpUrl, observation_id: str, api_root: str = APITerms.API.value,
                             headers=None):
    """
    Deletes an observation by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.OBSERVATIONS.value)
                   .with_resource_id(observation_id)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('DELETE')
                   .build())

    return api_request.make_request()


def list_all_procedures(server_addr: HttpUrl, api_root: str = APITerms.API.value, headers: dict = None):
    """
    Lists all procedures in the server at the default API endpoint
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.PROCEDURES.value)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('GET')
                   .build())

    return api_request.make_request()


def create_new_procedures(server_addr: HttpUrl, request_body: Union[str, dict], api_root: str = APITerms.API.value,
                          headers: dict = None):
    """
    Create a new procedure as defined by the request body
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.PROCEDURES.value)
                   .with_request_body(request_body)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('POST')
                   .build())
    print(api_request)
    return api_request.make_request()


def retrieve_procedure_by_id(server_addr: HttpUrl, procedure_id: str, api_root: str = APITerms.API.value,
                             headers: dict = None):
    """
    Retrieve a procedure by its ID
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.PROCEDURES.value)
                   .with_resource_id(procedure_id)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('GET')
                   .build())
    return api_request.make_request()


def update_procedure_by_id(server_addr: HttpUrl, procedure_id: str, request_body: Union[str, dict],
                           api_root: str = APITerms.API.value, headers: dict = None):
    """
    Update a procedure by its ID
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.PROCEDURES.value)
                   .with_resource_id(procedure_id)
                   .with_request_body(request_body)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('PUT')
                   .build())
    return api_request.make_request()


def delete_procedure_by_id(server_addr: HttpUrl, procedure_id: str, api_root: str = APITerms.API.value,
                           headers: dict = None):
    """
    Delete a procedure by its ID
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.PROCEDURES.value)
                   .with_resource_id(procedure_id)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('DELETE')
                   .build())
    return api_request.make_request()


def list_all_properties(server_addr: HttpUrl, api_root: str = APITerms.API.value):
    """
    List all properties
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.PROPERTIES.value)
                   .build_url_from_base()
                   .build())
    return api_request


def create_new_properties(server_addr: HttpUrl, request_body: dict, api_root: str = APITerms.API.value):
    """
    Create a new property as defined by the request body
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.PROPERTIES.value)
                   .with_request_body(request_body)
                   .build_url_from_base()
                   .build())
    return api_request


def retrieve_property_by_id(server_addr: HttpUrl, property_id: str, api_root: str = APITerms.API.value):
    """
    Retrieve a property by its ID
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.PROPERTIES.value)
                   .with_resource_id(property_id)
                   .build_url_from_base()
                   .build())
    return api_request


def update_property_by_id(server_addr: HttpUrl, property_id: str, request_body: dict,
                          api_root: str = APITerms.API.value):
    """
    Update a property by its ID
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.PROPERTIES.value)
                   .with_resource_id(property_id)
                   .with_request_body(request_body)
                   .build_url_from_base()
                   .build())
    return api_request


def delete_property_by_id(server_addr: HttpUrl, property_id: str, api_root: str = APITerms.API.value):
    """
    Delete a property by its ID
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.PROPERTIES.value)
                   .with_resource_id(property_id)
                   .build_url_from_base()
                   .build())
    return api_request


def list_all_sampling_features(server_addr: HttpUrl, api_root: str = APITerms.API.value, headers=None):
    """
    Lists all sampling features in the server at the default API endpoint
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.SAMPLING_FEATURES.value)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('GET')
                   .build())
    return api_request.make_request()


def list_sampling_features_of_system(server_addr: HttpUrl, system_id: str, api_root: str = APITerms.API.value,
                                     headers=None):
    """
    Lists all sampling features of a system by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.SYSTEMS.value)
                   .with_resource_id(system_id)
                   .for_sub_resource_type(APITerms.SAMPLING_FEATURES.value)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('GET')
                   .build())
    return api_request.make_request()


def create_new_sampling_features(server_addr: HttpUrl, system_id: str, request_body: Union[dict, str],
                                 api_root: str = APITerms.API.value, headers=None):
    """
    Create a new sampling feature as defined by the request body
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.SYSTEMS.value)
                   .with_resource_id(system_id)
                   .for_sub_resource_type(APITerms.SAMPLING_FEATURES.value)
                   .with_request_body(request_body)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('POST')
                   .build())
    return api_request.make_request()


def retrieve_sampling_feature_by_id(server_addr: HttpUrl, sampling_feature_id: str, api_root: str = APITerms.API.value,
                                    headers=None):
    """
    Retrieve a sampling feature by its ID
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.SAMPLING_FEATURES.value)
                   .with_resource_id(sampling_feature_id)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('GET')
                   .build())
    return api_request.make_request()


def update_sampling_feature_by_id(server_addr: HttpUrl, sampling_feature_id: str, request_body: Union[dict, str],
                                  api_root: str = APITerms.API.value, headers=None):
    """
    Update a sampling feature by its ID
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.SAMPLING_FEATURES.value)
                   .with_resource_id(sampling_feature_id)
                   .with_request_body(request_body)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('PUT')
                   .build())
    return api_request.make_request()


def delete_sampling_feature_by_id(server_addr: HttpUrl, sampling_feature_id: str, api_root: str = APITerms.API.value,
                                  headers=None):
    """
    Delete a sampling feature by its ID
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.SAMPLING_FEATURES.value)
                   .with_resource_id(sampling_feature_id)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('DELETE')
                   .build())
    return api_request.make_request()


def list_system_events(server_addr: HttpUrl, api_root: str = APITerms.API.value, headers: dict = None):
    """
    Lists all system events
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.SYSTEM_EVENTS.value)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('GET')
                   .build())
    return api_request.make_request()


def list_events_by_system_id(server_addr: HttpUrl, system_id: str, api_root: str = APITerms.API.value,
                             headers: dict = None):
    """
    Lists all events of a system
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.SYSTEMS.value)
                   .with_resource_id(system_id)
                   .for_sub_resource_type(APITerms.EVENTS.value)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('GET')
                   .build())
    return api_request.make_request()


def add_new_system_events(server_addr: HttpUrl, system_id: str, request_body: dict,
                          api_root: str = APITerms.API.value, headers: dict = None):
    """
    Adds a new system event to a system by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.SYSTEMS.value)
                   .with_resource_id(system_id)
                   .for_sub_resource_type(APITerms.EVENTS.value)
                   .with_request_body(request_body)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('POST')
                   .build())
    return api_request.make_request()


def retrieve_system_event_by_id(server_addr: HttpUrl, system_id: str, event_id: str,
                                api_root: str = APITerms.API.value, headers: dict = None):
    """
    Retrieves a system event by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.SYSTEMS.value)
                   .with_resource_id(system_id)
                   .for_sub_resource_type(APITerms.EVENTS.value)
                   .with_secondary_resource_id(event_id)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('GET')
                   .build())
    return api_request.make_request()


def update_system_event_by_id(server_addr: HttpUrl, system_id: str, event_id: str, request_body: dict,
                              api_root: str = APITerms.API.value, headers: dict = None):
    """
    Updates a system event by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.SYSTEMS.value)
                   .with_resource_id(system_id)
                   .for_sub_resource_type(APITerms.EVENTS.value)

                   .with_secondary_resource_id(event_id)
                   .with_request_body(request_body)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('PUT')
                   .build())
    return api_request.make_request()


def delete_system_event_by_id(server_addr: HttpUrl, system_id: str, event_id: str, api_root: str = APITerms.API.value,
                              headers: dict = None):
    """
    Deletes a system event by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.SYSTEMS.value)
                   .with_resource_id(system_id)
                   .for_sub_resource_type(APITerms.EVENTS.value)
                   .with_secondary_resource_id(event_id)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('DELETE')
                   .build())

    return api_request.make_request()


def list_system_history(server_addr: HttpUrl, system_id: str, api_root: str = APITerms.API.value, headers: dict = None):
    """
    Lists all history versions of a system
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.SYSTEMS.value)
                   .with_resource_id(system_id)
                   .for_resource_type(APITerms.HISTORY.value)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('GET')
                   .build())
    return api_request.make_request()


def retrieve_system_historical_description_by_id(server_addr: HttpUrl, system_id: str, history_id: str,
                                                 api_root: str = APITerms.API.value, headers: dict = None):
    """
    Retrieves a historical system description by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.SYSTEMS.value)
                   .with_resource_id(system_id)
                   .for_resource_type(APITerms.HISTORY.value)
                   .with_secondary_resource_id(history_id)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('GET')
                   .build())
    return api_request.make_request()


def update_system_historical_description(server_addr: HttpUrl, system_id: str, history_rev_id: str, request_body: dict,
                                         api_root: str = APITerms.API.value, headers: dict = None):
    """
    Updates a historical system description by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.SYSTEMS.value)
                   .with_resource_id(system_id)
                   .for_resource_type(APITerms.HISTORY.value)
                   .with_secondary_resource_id(history_rev_id)
                   .with_request_body(request_body)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('PUT')
                   .build())
    return api_request.make_request()


def delete_system_historical_description_by_id(server_addr: HttpUrl, system_id: str, history_rev_id: str,
                                               api_root: str = APITerms.API.value, headers: dict = None):
    """
    Deletes a historical system description by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.SYSTEMS.value)
                   .with_resource_id(system_id)
                   .for_resource_type(APITerms.HISTORY.value)
                   .with_secondary_resource_id(history_rev_id)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('DELETE')
                   .build())
    return api_request.make_request()


def list_all_systems(server_addr: HttpUrl, api_root: str = APITerms.API.value, headers: dict = None):
    """
    Lists all systems in the server at the default API endpoint
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.SYSTEMS.value)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('GET')
                   .build())

    return api_request.make_request()


def create_new_systems(server_addr: HttpUrl, request_body: Union[str, dict], api_root: str = APITerms.API.value,
                       uname: str = None,
                       pword: str = None, headers: dict = None):
    """
    Create a new system as defined by the request body
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.SYSTEMS.value)
                   .with_request_body(request_body)
                   .build_url_from_base()
                   .with_auth(uname, pword)
                   .with_headers(headers)
                   .with_request_method('POST')
                   .build())
    print(api_request.url)
    # resp = requests.post(api_request.url, data=api_request.body, headers=api_request.headers, auth=(uname, pword))
    resp = post_request(api_request.url, api_request.body, api_request.headers, api_request.auth)
    print(f'Create new system response: {resp}')
    return resp


def list_all_systems_in_collection(server_addr: HttpUrl, collection_id: str, api_root: str = APITerms.API.value):
    """
    NOTE: function may not be able to fully represent a request to the API at this time, as the test server lacks a few
    elements.
    Lists all systems in the server at the default API endpoint
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.COLLECTIONS.value)
                   .with_resource_id(collection_id)
                   # .for_sub_resource_type(APITerms.ITEMS.value)
                   .build_url_from_base()
                   .build())
    print(api_request.url)
    resp = requests.get(api_request.url, params=api_request.body, headers=api_request.headers)
    return resp.json()


def add_systems_to_collection(server_addr: HttpUrl, collection_id: str, uri_list: str,
                              api_root: str = APITerms.API.value):
    """
    Lists all systems in the server at the default API endpoint
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.COLLECTIONS.value)
                   .with_resource_id(collection_id)
                   .for_sub_resource_type(APITerms.ITEMS.value)
                   .with_request_body(uri_list)
                   .build_url_from_base()
                   .build())
    resp = requests.post(api_request.url, json=api_request.body, headers=api_request.headers)
    return resp.json()


def retrieve_system_by_id(server_addr: HttpUrl, system_id: str, api_root: str = APITerms.API.value):
    """
    Retrieves a system by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.SYSTEMS.value)
                   .with_resource_id(system_id)
                   .build_url_from_base()
                   .build())
    resp = requests.get(api_request.url, params=api_request.body, headers=api_request.headers)
    return resp.json()


def update_system_description(server_addr: HttpUrl, system_id: str, request_body: str,
                              api_root: str = APITerms.API.value, headers: dict = None):
    """
    Updates a system's description by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.SYSTEMS.value)
                   .with_resource_id(system_id)
                   .with_request_body(request_body)
                   .build_url_from_base()
                   .with_headers(headers)
                   .build())
    resp = requests.put(api_request.url, data=request_body, headers=api_request.headers)
    return resp


def delete_system_by_id(server_addr: HttpUrl, system_id: str, api_root: str = APITerms.API.value, headers: dict = None):
    """
    Deletes a system by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.SYSTEMS.value)
                   .with_resource_id(system_id)
                   .build_url_from_base()
                   .with_headers(headers)
                   .with_request_method('DELETE')
                   .build())
    return api_request.make_request()


def list_system_components(server_addr: HttpUrl, system_id: str, api_root: str = APITerms.API.value):
    """
    Lists all components of a system by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.SYSTEMS.value)
                   .with_resource_id(system_id)
                   .for_sub_resource_type(APITerms.COMPONENTS.value)
                   .build_url_from_base()
                   .build())
    print(api_request.url)
    resp = requests.get(api_request.url, params=api_request.body, headers=api_request.headers)
    return resp.json()


def add_system_components(server_addr: HttpUrl, system_id: str, request_body: dict,
                          api_root: str = APITerms.API.value):
    """
    Adds components to a system by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.SYSTEMS.value)
                   .with_resource_id(system_id)
                   .for_sub_resource_type(APITerms.COMPONENTS.value)
                   .with_request_body(request_body)
                   .build_url_from_base()
                   .build())
    resp = requests.post(api_request.url, params=api_request.body, headers=api_request.headers)
    return resp.json()


def list_deployments_of_system(server_addr: HttpUrl, system_id: str, api_root: str = APITerms.API.value):
    """
    Lists all deployments of a system by its id
    :return:
    """
    builder = ConnectedSystemsRequestBuilder()
    api_request = (builder.with_server_url(server_addr)
                   .with_api_root(api_root)
                   .for_resource_type(APITerms.SYSTEMS.value)
                   .with_resource_id(system_id)
                   .for_sub_resource_type(APITerms.DEPLOYMENTS.value)
                   .build_url_from_base()

                   .build())
    resp = requests.get(api_request.url, params=api_request.body, headers=api_request.headers)
    return resp.json()

# def list_sampling_features_of_system(server_addr: HttpUrl, system_id: str, api_root: str = APITerms.API.value):
#     """
#     Lists all sampling features of a system by its id
#     :return:
#     """
#     builder = ConnectedSystemsRequestBuilder()
#     api_request = (builder.with_server_url(server_addr)
#                    .with_api_root(api_root)
#                    .for_resource_type(APITerms.SYSTEMS.value)
#                    .with_resource_id(system_id)
#                    .for_sub_resource_type(APITerms.SAMPLING_FEATURES.value)
#                    .build_url_from_base()
#                    .build())
#     print(api_request.url)
#     resp = requests.get(api_request.url, params=api_request.body, headers=api_request.headers)
#     return resp.json()



from pydantic import BaseModel, HttpUrl


class NetworkProperties(BaseModel):
    endpoint_url: HttpUrl
    tls: bool = False
    stream_protocol: str = 'ws'
    mqtt_opts: dict = None
    mqtt_endpoint_url: HttpUrl = None
    connector_opts: dict = None



#   ==============================================================================
#   Copyright (c) 2024 Botts Innovative Research, Inc.
#   Date:  2024/5/28
#   Author:  Ian Patterson
#   Contact Email:  ian@botts-inc.com
#   ==============================================================================

import sys
import os
import websockets

from src.oshconnect import OSHConnect, Node
from timemanagement import TimePeriod, TimeInstant

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../src')))


class TestOSHConnect:
    TEST_PORT = 8282

    def test_time_period(self):
        tp = TimePeriod(start="2024-06-18T15:46:32Z", end="2024-06-18T20:00:00Z")
        assert tp is not None
        tps = tp.start
        tpe = tp.end
        assert isinstance(tps, TimeInstant)
        assert isinstance(tpe, TimeInstant)
        assert tps.epoch_time == TimeInstant.from_string("2024-06-18T15:46:32Z").epoch_time
        assert tpe.epoch_time == TimeInstant.from_string("2024-06-18T20:00:00Z").epoch_time

        tp = TimePeriod(start="now", end="2025-06-18T20:00:00Z")
        assert tp is not None
        assert tp.start == "now"
        assert tp.end.epoch_time == TimeInstant.from_string("2025-06-18T20:00:00Z").epoch_time

        tp = TimePeriod(start="2024-06-18T20:00:00Z", end="now")
        assert tp is not None
        assert tp.start.epoch_time == TimeInstant.from_string("2024-06-18T20:00:00Z").epoch_time
        assert tp.end == "now"

        # tp = TimePeriod(start="now", end="now")

    def test_oshconnect_create(self):
        app = OSHConnect(name="Test OSH Connect")
        assert app is not None
        assert app.get_name() == "Test OSH Connect"

    def test_oshconnect_add_node(self):
        app = OSHConnect(name="Test OSH Connect")
        node = Node(address="http://localhost", port=self.TEST_PORT, protocol="http", username="admin",
                    password="admin")
        # node.add_basicauth("admin", "admin")
        app.add_node(node)
        assert len(app._nodes) == 1
        assert app._nodes[0] == node

    def test_find_systems(self):
        app = OSHConnect(name="Test OSH Connect")
        node = Node(address="localhost", port=self.TEST_PORT, username="admin", password="admin", protocol="http")
        # node.add_basicauth("admin", "admin")
        app.add_node(node)
        app.discover_systems()
        print(f'Found systems: {app._systems}')
        # assert len(systems) == 1
        # assert systems[0] == node.get_api_endpoint()

    def test_oshconnect_find_datastreams(self):
        app = OSHConnect(name="Test OSH Connect")
        node = Node(address="localhost", port=self.TEST_PORT, username="admin", password="admin", protocol="http")
        app.add_node(node)
        app.discover_systems()

        app.discover_datastreams()
        assert len(app._datastreams) > 0

    async def test_obs_ws_stream(self):
        ds_url = (
            "ws://localhost:8282/sensorhub/api/datastreams/038q16egp1t0/observations?resultTime=latest"
            "/2026-01-01T12:00:00Z&f=application%2Fjson")

        # stream = requests.get(ds_url, stream=True, auth=('admin', 'admin'))
        async with websockets.connect(ds_url, extra_headers={'Authorization': 'Basic YWRtaW46YWRtaW4='}) as stream:
            async for message in stream:
                print(message)



from typing import Union

from pydantic import BaseModel, HttpUrl, Field

from .endpoints import Endpoint
from .request_wrappers import post_request, put_request, get_request, delete_request


class ConnectedSystemAPIRequest(BaseModel):
    url: HttpUrl = Field(None)
    body: Union[dict, str] = Field(None)
    params: dict = Field(None)
    request_method: str = Field('GET')
    headers: dict = Field(None)
    auth: Union[tuple, None] = Field(None)

    def make_request(self):
        match self.request_method:
            case 'GET':
                return get_request(self.url, self.params, self.headers, self.auth)
            case 'POST':
                print(f'POST request: {self}')
                return post_request(self.url, self.body, self.headers, self.auth)
            case 'PUT':
                print(f'PUT request: {self}')
                return put_request(self.url, self.body, self.headers, self.auth)
            case 'DELETE':
                print(f'DELETE request: {self}')
                return delete_request(self.url, self.params, self.headers, self.auth)
            case _:
                raise ValueError('Invalid request method')


class ConnectedSystemsRequestBuilder(BaseModel):
    api_request: ConnectedSystemAPIRequest = Field(default_factory=ConnectedSystemAPIRequest)
    base_url: HttpUrl = None
    endpoint: Endpoint = Field(default_factory=Endpoint)

    def with_api_url(self, url: HttpUrl):
        self.api_request.url = url
        return self

    def with_server_url(self, server_url: HttpUrl):
        self.base_url = server_url
        return self

    def build_url_from_base(self):
        """
        Builds the full API endpoint URL from the base URL and the endpoint parameters that have been previously
        provided.
        """
        self.api_request.url = f'{self.base_url}/{self.endpoint.create_endpoint()}'
        return self

    def with_api_root(self, api_root: str):
        """
        Optional: Set the API root for the request. This is useful if you want to use a different API root than the
        default one (api).
        :param api_root:
        :return:
        """
        self.endpoint.api_root = api_root
        return self

    def for_resource_type(self, resource_type: str):
        self.endpoint.base_resource = resource_type
        return self

    def with_resource_id(self, resource_id: str):
        self.endpoint.resource_id = resource_id
        return self

    def for_sub_resource_type(self, sub_resource_type: str):
        self.endpoint.sub_component = sub_resource_type
        return self

    def with_secondary_resource_id(self, resource_id: str):
        self.endpoint.secondary_resource_id = resource_id
        return self

    def with_request_body(self, request_body: str):
        self.api_request.body = request_body
        return self

    def with_request_method(self, request_method: str):
        self.api_request.request_method = request_method
        return self

    def with_headers(self, headers: dict = None):
        # TODO: ensure headers can default if excluded
        self.api_request.headers = headers
        return self

    def with_auth(self, uname: str, pword: str):
        self.api_request.auth = (uname, pword)
        return self

    def build(self):
        # convert endpoint to HttpUrl
        return self.api_request

    def reset(self):
        self.api_request = ConnectedSystemAPIRequest()
        self.endpoint = Endpoint()
        return self



#  =============================================================================
#  Copyright (c) 2025 Botts Innovative Research Inc.
#  Date: 2025/9/29
#  Author: Ian Patterson
#  Contact Email: ian@botts-inc.com
#  =============================================================================

from __future__ import annotations

import asyncio
import base64
import datetime
import json
import logging
import traceback
import uuid
from abc import ABC
from argparse import ArgumentError
from dataclasses import dataclass, field
from enum import Enum
from multiprocessing import Process
from multiprocessing.queues import Queue
from typing import TypeVar, Generic, Union
from uuid import UUID, uuid4
from collections import deque

from pydantic.v1.utils import to_lower_camel

from .csapi4py.constants import ContentTypes
from .schema_datamodels import JSONCommandSchema
from .csapi4py.mqtt import MQTTCommClient
from .csapi4py.constants import APIResourceTypes, ObservationFormat
from .csapi4py.default_api_helpers import APIHelper
from .encoding import JSONEncoding
from .resource_datamodels import ControlStreamResource
from .resource_datamodels import DatastreamResource, ObservationResource
from .resource_datamodels import SystemResource
from .schema_datamodels import SWEDatastreamRecordSchema
from .swe_components import DataRecordSchema
from .timemanagement import TimeInstant, TimePeriod, TimeUtils


@dataclass(kw_only=True)
class Endpoints:
    root: str = "sensorhub"
    sos: str = f"{root}/sos"
    connected_systems: str = f"{root}/api"


class Utilities:

    @staticmethod
    def convert_auth_to_base64(username: str, password: str) -> str:
        return base64.b64encode(f"{username}:{password}".encode()).decode()


class OSHClientSession:
    verify_ssl = True
    _streamables: dict[str, 'StreamableResource'] = None

    def __init__(self, base_url, *args, verify_ssl=True, **kwargs):
        # super().__init__(base_url, *args, **kwargs)
        self.verify_ssl = verify_ssl
        self._streamables = {}

    def connect_streamables(self):
        for streamable in self._streamables.values():
            streamable.start()

    def close_streamables(self):
        for streamable in self._streamables.values():
            streamable.stop()

    def register_streamable(self, streamable: StreamableResource):
        if self._streamables is None:
            self._streamables = {}
        self._streamables[streamable.get_streamable_id_str()] = streamable


class SessionManager:
    _session_tokens = None
    sessions: dict[str, OSHClientSession] = None

    def __init__(self, session_tokens: dict[str, str] = None):
        self._session_tokens = session_tokens
        self.sessions = {}

    def register_session(self, session_id, session: OSHClientSession) -> OSHClientSession:
        self.sessions[session_id] = session
        return session

    def unregister_session(self, session_id):
        session = self.sessions.pop(session_id)
        session.close()

    def get_session(self, session_id):
        return self.sessions.get(session_id, None)

    def start_session_streams(self, session_id):
        session = self.get_session(session_id)
        if session is None:
            raise ValueError(f"No session found for ID {session_id}")
        session.connect_streamables()

    def start_all_streams(self):
        for session in self.sessions.values():
            session.connect_streamables()


@dataclass(kw_only=True)
class Node:
    _id: str
    protocol: str
    address: str
    port: int
    server_root: str = 'sensorhub'
    endpoints: Endpoints
    is_secure: bool
    _basic_auth: bytes
    _api_helper: APIHelper
    _systems: list[System] = field(default_factory=list)
    _client_session: OSHClientSession
    _mqtt_client: MQTTCommClient
    _mqtt_port: int = 1883

    def __init__(self, protocol: str, address: str, port: int,
                 username: str = None, password: str = None, server_root: str = 'sensorhub',
                 session_manager: SessionManager = None,
                 **kwargs):
        self._id = f'node-{uuid.uuid4()}'
        self.protocol = protocol
        self.address = address
        self.server_root = server_root
        self.port = port
        self.is_secure = username is not None and password is not None
        if self.is_secure:
            self.add_basicauth(username, password)
        self.endpoints = Endpoints()
        self._api_helper = APIHelper(
            server_url=self.address,
            protocol=self.protocol,
            port=self.port,
            server_root=self.server_root,
            api_root='api', username=username,
            password=password)
        if self.is_secure:
            self._api_helper.user_auth = True
        self._systems = []
        if session_manager is not None:
            session_task = self.register_with_session_manager(session_manager)
            asyncio.gather(session_task)

        if kwargs.get('enable_mqtt'):
            if kwargs.get('mqtt_port') is not None:
                self._mqtt_port = kwargs.get('mqtt_port')
            self._mqtt_client = MQTTCommClient(url=self.address, port=self._mqtt_port,
                                               client_id_suffix=uuid.uuid4().hex, )
            self._mqtt_client.connect()
            self._mqtt_client.start()

    def get_id(self):
        return self._id

    def get_address(self):
        return self.address

    def get_port(self):
        return self.port

    def get_api_endpoint(self):
        return self._api_helper.get_api_root_url()

    def add_basicauth(self, username: str, password: str):
        if not self.is_secure:
            self.is_secure = True
        self._basic_auth = base64.b64encode(
            f"{username}:{password}".encode('utf-8'))

    def get_decoded_auth(self):
        return self._basic_auth.decode('utf-8')

    # def get_basicauth(self):
    #     return BasicAuth(self._api_helper.username, self._api_helper.password)

    def get_mqtt_client(self) -> MQTTCommClient:
        return self._mqtt_client

    def discover_systems(self):
        result = self._api_helper.retrieve_resource(APIResourceTypes.SYSTEM,
                                                    req_headers={})
        if result.ok:
            new_systems = []
            system_objs = result.json()['items']
            print(system_objs)
            for system_json in system_objs:
                print(system_json)
                system = SystemResource.model_validate(system_json)
                sys_obj = System(label=system.properties['name'],
                                 name=to_lower_camel(system.properties['name'].replace(" ", "_")),
                                 urn=system.properties['uid'], parent_node=self)

                self._systems.append(sys_obj)
                new_systems.append(sys_obj)
            return new_systems
        else:
            return None

    def add_new_system(self, system: System):
        system.set_parent_node(self)
        self._systems.append(system)

    def get_api_helper(self) -> APIHelper:
        return self._api_helper

    # System Management

    def add_system(self, system: System, target_node: Node, insert_resource: bool = False):
        """
        Add a system to the target node.
        :param system: System object
        :param target_node: Node object
        :param insert_resource: Whether to insert the system into the target node's server, default is False
        :return:
        """
        if insert_resource:
            system.insert_self()
        target_node.add_new_system(system)
        self._systems.append(system)
        return system

    def systems(self) -> list[System]:
        return self._systems

    def register_with_session_manager(self, session_manager: SessionManager):
        """
        Registers this node with the provided session manager, creating a new client session.
        :param session_manager: SessionManager instance
        """
        self._client_session = session_manager.register_session(self._id, OSHClientSession(
            base_url=self._api_helper.get_base_url()))

    def register_streamable(self, streamable: StreamableResource):
        if self._client_session is None:
            raise ValueError("Node is not registered with a SessionManager.")
        self._client_session.register_streamable(streamable)

    def get_session(self) -> OSHClientSession:
        return self._client_session


class Status(Enum):
    INITIALIZING = "initializing"
    INITIALIZED = "initialized"
    STARTING = "starting"
    STARTED = "started"
    STOPPING = "stopping"
    STOPPED = "stopped"


class StreamableModes(Enum):
    PUSH = "push"
    PULL = "pull"
    BIDIRECTIONAL = "bidirectional"


T = TypeVar('T', SystemResource, DatastreamResource, ControlStreamResource)


class StreamableResource(Generic[T], ABC):
    _id: UUID
    _resource_id: str
    _canonical_link: str
    _topic: str
    _status: str = Status.STOPPED.value
    ws_url: str
    _message_handler = None
    _parent_node: Node
    _underlying_resource: T
    _process: Process
    _msg_reader_queue: asyncio.Queue[Union[str, bytes, float, int]]
    _msg_writer_queue: asyncio.Queue[Union[str, bytes, float, int]]
    _inbound_deque: deque
    _outbound_deque: deque
    _mqtt_client: MQTTCommClient
    _parent_resource_id: str
    _connection_mode: StreamableModes = StreamableModes.PUSH.value

    def __init__(self, node: Node, connection_mode: StreamableModes = StreamableModes.PUSH.value):
        self._id = uuid4()
        self._parent_node = node
        self._parent_node.register_streamable(self)
        self._mqtt_client = self._parent_node.get_mqtt_client()
        self._connection_mode = connection_mode
        self._inbound_deque = deque()
        self._outbound_deque = deque()

    def get_streamable_id(self) -> UUID:
        return self._id

    def get_streamable_id_str(self) -> str:
        return self._id.hex

    def initialize(self):
        resource_type = None
        if isinstance(self._underlying_resource, SystemResource):
            resource_type = APIResourceTypes.SYSTEM
        elif isinstance(self._underlying_resource, DatastreamResource):
            resource_type = APIResourceTypes.DATASTREAM
        elif isinstance(self._underlying_resource, ControlStreamResource):
            resource_type = APIResourceTypes.CONTROL_CHANNEL
        if resource_type is None:
            raise ValueError(
                "Underlying resource must be set to either SystemResource or DatastreamResource before initialization.")
        # This needs to be implemented separately for each subclass
        res_id = getattr(self._underlying_resource, "ds_id", None) or getattr(self._underlying_resource, "cs_id", None)
        self.ws_url = self._parent_node.get_api_helper().construct_url(resource_type=resource_type,
                                                                       subresource_type=APIResourceTypes.OBSERVATION,
                                                                       resource_id=res_id,
                                                                       subresource_id=None)
        self._msg_reader_queue = asyncio.Queue()
        self._msg_writer_queue = asyncio.Queue()
        self.init_mqtt()
        self._status = Status.INITIALIZED.value

    def start(self):
        if self._status != Status.INITIALIZED.value:
            logging.warning(f"Streamable resource {self._id} not initialized. Call initialize() first.")
            return
        self._status = Status.STARTING.value
        self._status = Status.STARTED.value

    async def stream(self):
        session = self._parent_node.get_session()

        try:
            async with session.ws_connect(self.ws_url, auth=self._parent_node.get_basicauth()) as ws:
                logging.info(f"Streamable resource {self._id} started.")
                read_task = asyncio.create_task(self._read_from_ws(ws))
                write_task = asyncio.create_task(self._write_to_ws(ws))
                await asyncio.gather(read_task, write_task)
        except Exception as e:
            logging.error(f"Error in streamable resource {self._id}: {e}")
            logging.error(traceback.format_exc())

    def init_mqtt(self):
        if self._mqtt_client is None:
            logging.warning(f"No MQTT client configured for streamable resource {self._id}.")
            return

        self._mqtt_client.set_on_subscribe(self._default_on_subscribe)

        # self.get_mqtt_topic()

    def _default_on_subscribe(self, client, userdata, mid, granted_qos, properties):
        print("OSH Subscribed: " + str(mid) + " " + str(granted_qos))

    def get_mqtt_topic(self, subresource: APIResourceTypes | None = None):
        """
        Retrieves the MQTT topic for this streamable resource based on its underlying resource type. By default, the topic
        is actually for listening to subresources of a default type
        :param subresource : Optional subresource type to get the topic for, defaults to None
        """
        resource_type = None
        parent_res_type = None
        parent_id = None

        if isinstance(self._underlying_resource, ControlStreamResource):
            parent_res_type = APIResourceTypes.CONTROL_CHANNEL
            parent_id = self._resource_id

            match subresource:
                case APIResourceTypes.COMMAND:
                    resource_type = APIResourceTypes.COMMAND
                case APIResourceTypes.STATUS:
                    resource_type = APIResourceTypes.STATUS

        elif isinstance(self._underlying_resource, DatastreamResource):
            parent_res_type = APIResourceTypes.DATASTREAM
            resource_type = APIResourceTypes.OBSERVATION
            parent_id = self._resource_id

        elif isinstance(self._underlying_resource, SystemResource):
            match subresource:
                case APIResourceTypes.DATASTREAM:
                    resource_type = APIResourceTypes.DATASTREAM
                    parent_res_type = APIResourceTypes.SYSTEM
                    parent_id = self._resource_id
                case APIResourceTypes.CONTROL_CHANNEL:
                    resource_type = APIResourceTypes.CONTROL_CHANNEL
                    parent_res_type = APIResourceTypes.SYSTEM
                    parent_id = self._resource_id
                case None:
                    resource_type = APIResourceTypes.SYSTEM
                    parent_res_type = None
                    parent_id = None
                case _:
                    raise ValueError(f"Unsupported subresource type {subresource} for SystemResource.")

        topic = self._parent_node.get_api_helper().get_mqtt_topic(subresource_type=resource_type,
                                                                  resource_id=parent_id,
                                                                  resource_type=parent_res_type)
        return topic

    async def _read_from_ws(self, ws):
        async for msg in ws:
            self._message_handler(ws, msg)

    async def _write_to_ws(self, ws):
        while self._status is Status.STARTED.value:
            try:
                msg = self._msg_writer_queue.get_nowait()
                await ws.send_bytes(msg)
            except asyncio.QueueEmpty:
                await asyncio.sleep(0.05)

    def stop(self):
        # It would be nicer to join() here once we have cleaner shutdown logic in place to avoid corrupting processes
        # that are writing to streams or that need to manage authentication state
        self._status = "stopping"
        self._process.terminate()
        self._status = "stopped"

    def set_parent_node(self, node: Node):
        self._parent_node = node

    def get_parent_node(self) -> Node:
        return self._parent_node

    def set_parent_resource_id(self, res_id: str):
        self._parent_resource_id = res_id

    def get_parent_resource_id(self) -> str:
        return self._parent_resource_id

    def set_connection_mode(self, connection_mode: StreamableModes):
        self._connection_mode = connection_mode

    def poll(self):
        pass

    def fetch(self, time_period: TimePeriod):
        pass

    def get_msg_reader_queue(self) -> Queue:
        """
        Returns the message queue for this streamable resource. In cases where a custom message handler is used this is
        not guaranteed to return anything or provided a queue with data.
        :return: Queue object
        """
        return self._msg_reader_queue

    def get_msg_writer_queue(self) -> Queue:
        """
        Returns the message queue for writing messages to this streamable resource.
        :return: Queue object
        """
        return self._msg_writer_queue

    def get_underlying_resource(self) -> T:
        return self._underlying_resource

    def get_internal_id(self) -> UUID:
        return self._id

    def insert_data(self, data: dict):
        """ Naively inserts data into the message writer queue to be sent over the WebSocket connection.
            No Checks are performed to ensure the data is valid for the underlying resource.
            :param data: Data to be sent, typically bytes or str
        """
        print(f"Inserting data into message writer queue: {data}")
        data_bytes = json.dumps(data).encode("utf-8") if isinstance(data, dict) else data
        self._msg_writer_queue.put_nowait(data_bytes)

    def subscribe_mqtt(self, topic: str, qos: int = 0):
        if self._mqtt_client is None:
            logging.warning(f"No MQTT client configured for streamable resource {self._id}.")
            return
        self._mqtt_client.subscribe(topic, qos=qos, msg_callback=self._mqtt_sub_callback)

    def _publish_mqtt(self, topic, payload):
        if self._mqtt_client is None:
            logging.warning(f"No MQTT client configured for streamable resource {self._id}.")
            return
        print(f'Publishing to MQTT topic {topic}: {payload}')
        self._mqtt_client.publish(topic, payload, qos=0)

    async def _write_to_mqtt(self):
        while self._status is Status.STARTED.value:
            try:
                msg = self._outbound_deque.popleft()
                print(f"Popped message: {msg}, attempting to publish...")
                self._publish_mqtt(self._topic, msg)
            except IndexError:
                await asyncio.sleep(0.05)
            except Exception as e:
                print(f"Error in Write To MQTT {self._id}: {e}")
                print(traceback.format_exc())
        if self._status is Status.STOPPED.value:
            print("MQTT write task stopping as streamable resource is stopped.")

    def publish(self, payload, topic: str = None):
        """
        Publishes data to the MQTT topic associated with this streamable resource.
        :param payload: Data to be published, subclass should determine specifically allowed types
        :param topic: Specific implementation determines the topic from the provided string, if None the default topic is used
        """
        self._publish_mqtt(self._topic, payload)

    def subscribe(self, topic=None, callback=None, qos=0):
        """
        Subscribes to the MQTT topic associated with this streamable resource.
        :param topic: Specific implementation determines the topic from the provided string, if None the default topic is used
        :param callback: Optional callback function to handle incoming messages, if None the default handler is used
        :param qos: Quality of Service level for the subscription, default is 0
        """
        t = None

        if topic is None:
            t = self._topic
        else:
            raise ArgumentError("Invalid topic provided, must be None to use default topic.")

        if callback is None:
            self._mqtt_client.subscribe(t, qos=qos, msg_callback=self._mqtt_sub_callback)
        else:
            self._mqtt_client.subscribe(t, qos=qos, msg_callback=callback)

    def _mqtt_sub_callback(self, client, userdata, msg):
        print(f"Received MQTT message on topic {msg.topic}: {msg.payload}")
        # Appends to right of deque
        self._inbound_deque.append(msg.payload)

    def get_inbound_deque(self):
        return self._inbound_deque

    def get_outbound_deque(self):
        return self._outbound_deque


class System(StreamableResource[SystemResource]):
    name: str
    label: str
    datastreams: list[Datastream]
    control_channels: list[ControlStream]
    description: str
    urn: str
    _parent_node: Node

    def __init__(self, name: str, label: str, urn: str, parent_node: Node, **kwargs):
        """
        :param name: The machine-accessible name of the system
        :param label: The human-readable label of the system
        :param urn: The URN of the system, typically formed as such: 'urn:general_identifier:specific_identifier:more_specific_identifier'
        :param kwargs:
            - 'description': A description of the system
        """
        super().__init__(node=parent_node)
        self.name = name
        self.label = label
        self.datastreams = []
        self.control_channels = []
        self.urn = urn
        if kwargs.get('resource_id'):
            self._resource_id = kwargs['resource_id']
        if kwargs.get('description'):
            self.description = kwargs['description']

        self._underlying_resource = self.to_system_resource()

    def discover_datastreams(self) -> list[DatastreamResource]:
        res = self._parent_node.get_api_helper().get_resource(APIResourceTypes.SYSTEM, self._resource_id,
                                                              APIResourceTypes.DATASTREAM)
        datastream_json = res.json()['items']
        ds_resources = []

        for ds in datastream_json:
            datastream_objs = DatastreamResource.model_validate(ds)
            ds_resources.append(datastream_objs)

        return ds_resources

    def discover_controlstreams(self) -> list[ControlStreamResource]:
        res = self._parent_node.get_api_helper().get_resource(APIResourceTypes.SYSTEM, self._resource_id,
                                                              APIResourceTypes.CONTROL_CHANNEL)
        controlstream_json = res.json()['items']
        cs_resources = []

        for cs in controlstream_json:
            controlstream_objs = ControlStreamResource.model_validate(cs)
            cs_resources.append(controlstream_objs)

        return cs_resources

    @staticmethod
    def from_system_resource(system_resource: SystemResource, parent_node: Node) -> System:
        other_props = system_resource.model_dump()
        print(f'Props of SystemResource: {other_props}')

        # case 1: has properties a la geojson
        if 'properties' in other_props:
            new_system = System(name=other_props['properties']['name'],
                                label=other_props['properties']['name'],
                                urn=other_props['properties']['uid'],
                                resource_id=system_resource.system_id, parent_node=parent_node)
        else:
            new_system = System(name=system_resource.name,
                                label=system_resource.label, urn=system_resource.urn,
                                resource_id=system_resource.system_id, parent_node=parent_node)

        new_system.set_system_resource(system_resource)
        return new_system

    def to_system_resource(self) -> SystemResource:
        resource = SystemResource(uid=self.urn, label=self.name, feature_type='PhysicalSystem')

        if len(self.datastreams) > 0:
            resource.outputs = [ds.get_underlying_resource() for ds in self.datastreams]

        # if len(self.control_channels) > 0:
        #     resource.inputs = [cc.to_resource() for cc in self.control_channels]
        return resource

    def set_system_resource(self, sys_resource: SystemResource):
        self._underlying_resource = sys_resource

    def get_system_resource(self) -> SystemResource:
        return self._underlying_resource

    def add_insert_datastream(self, datarecord_schema: DataRecordSchema):
        """
        Adds a datastream to the system while also inserting it into the system's parent node via HTTP POST.
        :param datarecord_schema: DataRecordSchema to be used to define the datastream
        :return:
        """
        print(f'Adding datastream: {datarecord_schema.model_dump_json(exclude_none=True, by_alias=True)}')
        # Make the request to add the datastream
        # if successful, add the datastream to the system
        datastream_schema = SWEDatastreamRecordSchema(record_schema=datarecord_schema,
                                                      obs_format='application/swe+json',
                                                      encoding=JSONEncoding())
        datastream_resource = DatastreamResource(ds_id="default", name=datarecord_schema.label,
                                                 output_name=datarecord_schema.label,
                                                 record_schema=datastream_schema,
                                                 valid_time=TimePeriod(start=TimeInstant.now_as_time_instant(),
                                                                       end=TimeInstant(utc_time=TimeUtils.to_utc_time(
                                                                           "2026-12-31T00:00:00Z"))))

        api = self._parent_node.get_api_helper()
        print(
            f'Attempting to create datastream: {datastream_resource.model_dump(by_alias=True, exclude_none=True)}')
        res = api.create_resource(APIResourceTypes.DATASTREAM,
                                  datastream_resource.model_dump_json(by_alias=True, exclude_none=True),
                                  req_headers={
                                      'Content-Type': ContentTypes.JSON.value
                                  }, parent_res_id=self._resource_id)

        if res.ok:
            datastream_id = res.headers['Location'].split('/')[-1]
            print(f'Resource Location: {datastream_id}')
            datastream_resource.ds_id = datastream_id
        else:
            raise Exception(f'Failed to create datastream: {datastream_resource.name}')

        new_ds = Datastream(self._parent_node, datastream_resource)
        new_ds.set_parent_resource_id(self._underlying_resource.system_id)
        self.datastreams.append(new_ds)
        return new_ds

    def add_and_insert_control_stream(self, control_stream_record_schema: DataRecordSchema, input_name: str = None,
                                      valid_time: TimePeriod = None) -> ControlStream:
        """
        Accepts a DataRecordSchema and creates a JSON encoded schema structure ControlStreamResource, which is inserted
        into the parent system via the host node.
        :param control_stream_record_schema: DataRecordSchema to be used for the control stream
        :param input_name: Name of the input, if None the label of the schema is converted to lower and stripped of whitespace
        :return: ControlStream object added to the system
        """
        input_name_checked = input_name if input_name is not None else control_stream_record_schema.label.lower().replace(
            ' ', '')

        now = datetime.datetime.now()
        future_time = now.replace(year=now.year + 1)
        future_str = future_time.strftime("%Y-%m-%dT%H:%M:%SZ")

        valid_time_checked = valid_time if valid_time else TimePeriod(start=TimeInstant.now_as_time_instant(),
                                                                      end=TimeInstant(
                                                                          utc_time=TimeUtils.to_utc_time(future_str)))

        command_schema = JSONCommandSchema(command_format=ObservationFormat.SWE_JSON.value,
                                           params_schema=control_stream_record_schema)
        control_stream_resource = ControlStreamResource(name=control_stream_record_schema.label,
                                                        input_name=input_name_checked,
                                                        command_schema=command_schema,
                                                        validTime=valid_time_checked)
        api = self._parent_node.get_api_helper()
        res = api.create_resource(APIResourceTypes.CONTROL_CHANNEL,
                                  control_stream_resource.model_dump_json(by_alias=True, exclude_none=True),
                                  req_headers={
                                      'Content-Type': 'application/json'
                                  }, parent_res_id=self._resource_id)

        if res.ok:
            control_channel_id = res.headers['Location'].split('/')[-1]
            print(f'Control Stream Resource Location: {control_channel_id}')
            control_stream_resource.cs_id = control_channel_id
        else:
            raise Exception(f'Failed to create control stream: {control_stream_resource.name}')

        new_cs = ControlStream(node=self._parent_node, controlstream_resource=control_stream_resource)
        new_cs.set_parent_resource_id(self._underlying_resource.system_id)
        self.control_channels.append(new_cs)
        return new_cs

    def insert_self(self):
        res = self._parent_node.get_api_helper().create_resource(
            APIResourceTypes.SYSTEM, self.to_system_resource().model_dump_json(by_alias=True, exclude_none=True),
            req_headers={
                'Content-Type': 'application/sml+json'
            })

        if res.ok:
            location = res.headers['Location']
            sys_id = location.split('/')[-1]
            self._resource_id = sys_id
            print(f'Created system: {self._resource_id}')

    def retrieve_resource(self):
        if self._resource_id is None:
            return None
        res = self._parent_node.get_api_helper().retrieve_resource(res_type=APIResourceTypes.SYSTEM,
                                                                   res_id=self._resource_id)
        if res.ok:
            system_json = res.json()
            print(system_json)
            system_resource = SystemResource.model_validate(system_json)
            print(f'System Resource: {system_resource}')
            self._underlying_resource = system_resource
            return None


class Datastream(StreamableResource[DatastreamResource]):
    should_poll: bool

    def __init__(self, parent_node: Node = None, datastream_resource: DatastreamResource = None):
        super().__init__(node=parent_node)
        self._underlying_resource = datastream_resource
        self._resource_id = datastream_resource.ds_id

    def get_id(self):
        return self._underlying_resource.ds_id

    @staticmethod
    def from_resource(ds_resource: DatastreamResource, parent_node: Node):
        new_ds = Datastream(parent_node=parent_node, datastream_resource=ds_resource)
        return new_ds

    def set_resource(self, resource: DatastreamResource):
        self._underlying_resource = resource

    def get_resource(self) -> DatastreamResource:
        return self._underlying_resource

    def create_observation(self, obs_data: dict):
        obs = ObservationResource(result=obs_data, result_time=TimeInstant.now_as_time_instant())
        # Validate against the schema
        if self._underlying_resource.record_schema is not None:
            obs.validate_against_schema(self._underlying_resource.record_schema)
        return obs

    def insert_observation_dict(self, obs_data: dict):
        res = self._parent_node.get_api_helper().create_resource(APIResourceTypes.OBSERVATION, obs_data,
                                                                 parent_res_id=self._resource_id,
                                                                 req_headers={'Content-Type': 'application/json'})
        if res.ok:
            obs_id = res.headers['Location'].split('/')[-1]
            print(f'Inserted observation: {obs_id}')
            return id
        else:
            raise Exception(f'Failed to insert observation: {res.text}')

    def start(self):
        super().start()
        if self._mqtt_client is not None:
            # self._mqtt_client.connect()

            if self._connection_mode is StreamableModes.PULL or self._connection_mode is StreamableModes.BIDIRECTIONAL:
                self._mqtt_client.subscribe(self._topic, msg_callback=self._mqtt_sub_callback)
            else:
                try:
                    loop = asyncio.get_event_loop()
                    loop.create_task(self._write_to_mqtt())
                except Exception as e:
                    # TODO: Use logging instead of print
                    print(traceback.format_exc())
                    print(f"Error starting MQTT write task: {e}")

            # self._mqtt_client.start()

    def init_mqtt(self):
        super().init_mqtt()
        self._topic = self.get_mqtt_topic(subresource=APIResourceTypes.OBSERVATION)

    def _queue_push(self, msg):
        print(f'Pushing message to reader queue: {msg}')
        self._msg_writer_queue.put_nowait(msg)
        print(f'Queue size is now: {self._msg_writer_queue.qsize()}')

    def _queue_pop(self):
        return self._msg_reader_queue.get_nowait()

    def insert(self, data: dict):
        # self._queue_push(data)
        encoded = json.dumps(data).encode('utf-8')
        self._publish_mqtt(self._topic, encoded)


class ControlStream(StreamableResource[ControlStreamResource]):
    _status_topic: str
    _inbound_status_deque: deque
    _outbound_status_deque: deque

    def __init__(self, node: Node = None, controlstream_resource: ControlStreamResource = None):
        super().__init__(node=node)
        self._underlying_resource = controlstream_resource
        self._inbound_status_deque = deque()
        self._outbound_status_deque = deque()
        self._resource_id = controlstream_resource.cs_id
        # Always make sure this is set after the resource ids are set
        self._status_topic = self.get_mqtt_status_topic()

    def add_underlying_resource(self, resource: ControlStreamResource):
        self._underlying_resource = resource

    def init_mqtt(self):
        super().init_mqtt()
        self._topic = self.get_mqtt_topic(subresource=APIResourceTypes.COMMAND)

    def get_mqtt_status_topic(self):
        return self.get_mqtt_topic(subresource=APIResourceTypes.STATUS)

    def start(self):
        super().start()
        if self._mqtt_client is not None:
            if self._connection_mode is StreamableModes.PULL or self._connection_mode is StreamableModes.BIDIRECTIONAL:
                # Subs to command topic by default
                self._mqtt_client.subscribe(self._topic, msg_callback=self._mqtt_sub_callback)
            else:
                try:
                    loop = asyncio.get_event_loop()
                    loop.create_task(self._write_to_mqtt())
                except Exception as e:
                    print(traceback.format_exc())
                    print(f"Error starting MQTT write task: {e}")

    def get_inbound_deque(self):
        return self._inbound_deque

    def get_outbound_deque(self):
        return self._outbound_deque

    def get_status_deque_inbound(self):
        return self._inbound_status_deque

    def get_status_deque_outbound(self):
        return self._outbound_status_deque

    def publish_command(self, payload):
        self.publish(payload, topic=APIResourceTypes.COMMAND.value)

    def publish_status(self, payload):
        self.publish(payload, topic=APIResourceTypes.STATUS.value)

    def publish(self, payload, topic: str = 'command'):
        """
        Publishes data to the MQTT topic associated with this control stream resource.
        :param payload: Data to be published, subclass should determine specifically allowed types
        :param topic: Specific implementation determines the topic from the provided string
        """

        if topic == APIResourceTypes.COMMAND.value:
            self._publish_mqtt(self._topic, payload)
        elif topic == APIResourceTypes.STATUS.value:
            self._publish_mqtt(self._status_topic, payload)
        else:
            raise ValueError(f"Unsupported topic type {topic} for ControlStream publish().")

    def subscribe(self, topic=None, callback=None, qos=0):
        """
        Subscribes to the MQTT topic associated with this control stream resource.
        :param topic: Specific implementation determines the topic from the provided string
        :param callback: Optional callback function to handle incoming messages, if None the default handler is used
        :param qos: Quality of Service level for the subscription, default is 0
        """

        t = None

        if topic is None or topic == APIResourceTypes.COMMAND.value:
            t = self._topic
        elif topic == APIResourceTypes.STATUS.value:
            t = self._status_topic
        else:
            raise ArgumentError(f"Invalid topic provided {topic}, must be None or one of 'command' or 'status'.")

        if callback is None:
            self._mqtt_client.subscribe(t, qos=qos, msg_callback=self._mqtt_sub_callback)
        else:
            self._mqtt_client.subscribe(t, qos=qos, msg_callback=callback)



from typing import Union

from pydantic import BaseModel, HttpUrl, Field, model_serializer, RootModel, SerializeAsAny

from .constants import DatastreamResultTypes
from oshconnect.datamodels.datastreams import DatastreamSchema
from oshconnect.datamodels.geometry import Geometry
from .sensor_ml.sml import TypeOf


# TODO: Consider some sort of Abstract Base Class for all valid request bodies to inherit from to reduce the complexity
#  of the final request body.

class GeoJSONBody(BaseModel):
    type: str
    id: str
    properties: dict = None
    geometry: Geometry = None
    bbox: list = None
    links: list = None


class SmlJSONBody(BaseModel):
    object_type: str = Field(None, serialization_alias='type')
    id: str = Field(None)
    description: str = Field(None)
    unique_id: str = Field(..., serialization_alias='uniqueId')
    label: str = Field(...)
    lang: str = None
    keywords: list = None
    identifiers: list = None
    classifiers: list = None
    valid_time: list = Field(None, serialization_alias='validTime')
    security_constraints: list = Field(None, serialization_alias='securityConstraints')
    legal_constraints: list = Field(None, serialization_alias='legalConstraints')
    characteristics: list = None
    capabilities: list = None
    contacts: list = None
    documents: list = None
    history: list = None
    definition: HttpUrl = None
    type_of: TypeOf = Field(None, serialization_alias='typeOf')
    configuration: HttpUrl = None
    features_of_interest: list = Field(None, serialization_alias='featuresOfInterest')
    inputs: list = None
    outputs: list = None
    parameters: list = None
    modes: list = None
    method: str = None
    position: list = None
    links: list = Field(None)


class OMJSONBody(BaseModel):
    datastream_id: str = Field(None, alias="datastream@id")
    foi_id: str = Field(None, alias="foi@id")
    phenomenon_time: str = Field(None, alias="phenomenonTime")
    result_time: str = Field(None, alias="resultTime")
    parameters: list = Field(None)
    result: dict = Field(None)
    result_links: list = Field(None, alias="result@links")


class DatastreamBodyJSON(BaseModel):
    """
    NOTES: though the spec does not require that outputName, and schema be present, they are required for the
    implementation of the API present on OSH
    """
    id: str = Field(None)
    name: str = Field(...)
    description: str = Field(None)
    deployment: HttpUrl = Field(None, serialization_alias='deployment@link')
    ultimate_feature_of_interest: HttpUrl = Field(None, serialization_alias='featureOfInterest@link')
    sampling_feature: HttpUrl = Field(None, serialization_alias='samplingFeature@link')
    valid_time: list = Field(None, serialization_alias='validTime')
    output_name: str = Field(..., serialization_alias='outputName')
    phenomenon_time_interval: str = Field(None, serialization_alias='phenomenonTimeInterval')
    result_time_interval: str = Field(None, serialization_alias='resultTimeInterval')
    result_type: DatastreamResultTypes = Field(None, serialization_alias='resultType')
    links: list = Field(None)
    datastream_schema: SerializeAsAny[DatastreamSchema] = Field(..., serialization_alias='schema')


class RequestBody(BaseModel):
    """
    Wrapper class to support different request json structures
    """
    json_structure: Union[GeoJSONBody, SmlJSONBody, OMJSONBody, DatastreamSchema] = Field(...,
                                                                                          serialization_alias='json')
    test_extra: str = Field("Hello, I am test", serialization_alias='testExtra')

    @model_serializer
    def ser_model(self):
        print("Serializing model...")
        return self.json_structure


class RequestBodyList(RootModel):
    root: list[Union[GeoJSONBody, SmlJSONBody, OMJSONBody, DatastreamSchema]] = Field(...)



from typing import Union

import requests
from pydantic import HttpUrl


def get_request(url: HttpUrl, params: dict = None, headers: dict = None, auth: tuple = None):
    """
    Sends a GET request to the provided URL with the given parameters and headers
    :param url:
    :param params:
    :param headers:
    :param auth:
    :return: the response of the request
    """
    return requests.get(url, params=params, headers=headers, auth=auth)


def post_request(url: HttpUrl, body: Union[str, dict] = None, headers: dict = None, auth: tuple = None):
    """
    Sends a POST request to the provided URL with the given content and headers
    :param url:
    :param content_json:
    :param headers:
    :param auth:
    :return: the response of the request
    """
    if isinstance(body, str):
        return requests.post(url, data=body, headers=headers, auth=auth)
    else:
        return requests.post(url, json=body, headers=headers, auth=auth)


def put_request(url: HttpUrl, body: Union[str, dict] = None, headers: dict = None, auth: tuple = None):
    """
    Sends a PUT request to the provided URL with the given content and headers
    :param url:
    :param content_json:
    :param headers:
    :param auth:
    :return: the response of the request
    """
    if isinstance(body, str):
        return requests.put(url, data=body, headers=headers, auth=auth)
    else:
        return requests.put(url, json=body, headers=headers, auth=auth)


def delete_request(url: HttpUrl, params: dict = None, headers: dict = None, auth: tuple = None):
    """
    Sends a DELETE request to the provided URL with the given parameters and headers
    :param url:
    :param params:
    :param headers:
    :param auth:
    :return: the response of the request
    """
    return requests.delete(url, params=params, headers=headers, auth=auth)



#  =============================================================================
#  Copyright (c) 2025 Botts Innovative Research Inc.
#  Date: 2025/10/6
#  Author: Ian Patterson
#  Contact Email: ian@botts-inc.com
#  =============================================================================
import collections
from typing import Any
from uuid import UUID
from abc import ABC


class Event(ABC):
    """
    A base class for events in the event bus system.
    """
    id: UUID
    topic: str
    payload: Any

    def __init__(self, id: UUID, topic: str, payload: Any):
        self.id = id
        self.topic = topic
        self.payload = payload


class EventBus(ABC):
    """
    A base class for an event bus system.
    """
    _deque: collections.deque



#  =============================================================================
#  Copyright (c) 2025 Botts Innovative Research Inc.
#  Date: 2025/9/30
#  Author: Ian Patterson
#  Contact Email: ian@botts-inc.com
#  =============================================================================

from __future__ import annotations

from numbers import Real
from typing import Union, Any

from pydantic import BaseModel, Field, field_validator, SerializeAsAny

from .csapi4py.constants import GeometryTypes
from .api_utils import UCUMCode, URI
from .geometry import Geometry

"""
 NOTE: The following classes are used to represent the Record Schemas that are required for use with Datastreams
The names are likely to change to include a "Schema" suffix to differentiate them from the actual data structures.
The current scope of the project likely excludes conversion from received data to actual SWE Common data structures,
in the event this is added it will most likely be in a separate module as those structures have use cases outside of
the API solely
"""


# TODO: Add field validators that are missing
# TODO: valid string fields that are intended to represent time/date values
# TODO: Validate places where string fields are not allowed to be empty


class AnyComponentSchema(BaseModel):
    type: str = Field(...)
    id: str = Field(None)
    label: str = Field(None)
    description: str = Field(None)
    updatable: bool = Field(False)
    optional: bool = Field(False)
    definition: str = Field(None)


class DataRecordSchema(AnyComponentSchema):
    type: str = "DataRecord"
    fields: SerializeAsAny[list[AnyComponentSchema]] = Field(...)


class VectorSchema(AnyComponentSchema):
    label: str = Field(...)
    name: str = Field(...)
    type: str = "Vector"
    definition: str = Field(...)
    reference_frame: str = Field(...)
    local_frame: str = Field(None)
    # TODO: VERIFY might need to be moved further down when these are defined
    coordinates: SerializeAsAny[Union[list[CountSchema], list[QuantitySchema], list[TimeSchema]]] = Field(...)


class DataArraySchema(AnyComponentSchema):
    type: str = "DataArray"
    name: str = Field(...)
    element_count: dict | str | CountSchema = Field(..., serialization_alias='elementCount')  # Should type of Count
    element_type: SerializeAsAny[list[AnyComponentSchema]] = Field(..., serialization_alias='elementType')
    encoding: str = Field(...)  # TODO: implement an encodings class
    values: list = Field(None)


class MatrixSchema(AnyComponentSchema):
    type: str = "Matrix"
    element_count: dict | str | CountSchema = Field(..., serialization_alias='elementCount')  # Should be type of Count
    element_type: SerializeAsAny[list[AnyComponentSchema]] = Field(..., serialization_alias='elementType')
    encoding: str = Field(...)  # TODO: implement an encodings class
    values: list = Field(None)
    reference_frame: str = Field(None)
    local_frame: str = Field(None)


class DataChoiceSchema(AnyComponentSchema):
    type: str = "DataChoice"
    updatable: bool = Field(False)
    optional: bool = Field(False)
    choice_value: CategorySchema = Field(..., serialization_alias='choiceValue')  # TODO: Might be called "choiceValues"
    items: SerializeAsAny[list[AnyComponentSchema]] = Field(...)


class GeometrySchema(AnyComponentSchema):
    label: str = Field(...)
    type: str = "Geometry"
    updatable: bool = Field(False)
    optional: bool = Field(False)
    definition: str = Field(...)
    constraint: dict = Field(default_factory=lambda: {
        'geomTypes': [
            GeometryTypes.POINT.value,
            GeometryTypes.LINESTRING.value,
            GeometryTypes.POLYGON.value,
            GeometryTypes.MULTI_POINT.value,
            GeometryTypes.MULTI_LINESTRING.value,
            GeometryTypes.MULTI_POLYGON.value
        ]
    })
    nil_values: list = Field(None, serialization_alias='nilValues')
    srs: str = Field(...)
    value: Geometry = Field(None)


class AnySimpleComponentSchema(AnyComponentSchema):
    label: str = Field(...)
    description: str = Field(None)
    type: str = Field(...)
    updatable: bool = Field(False)
    optional: bool = Field(False)
    definition: str = Field(...)
    reference_frame: str = Field(None, serialization_alias='referenceFrame')
    axis_id: str = Field(None, serialization_alias='axisID')
    quality: Union[list[QuantitySchema], list[QuantityRangeSchema], list[CategorySchema], list[TextSchema]] = Field(
        None)  # TODO: Union[Quantity, QuantityRange, Category, Text]
    nil_values: list = Field(None, serialization_alias='nilValues')
    constraint: Any = Field(None)
    value: Any = Field(None)
    name: str = Field(...)


class AnyScalarComponentSchema(AnySimpleComponentSchema):
    """
    A base class for all scalar components. The structure is essentially that of AnySimpleComponent
    """
    pass


class BooleanSchema(AnyScalarComponentSchema):
    type: str = "Boolean"
    value: bool = Field(None)


class CountSchema(AnyScalarComponentSchema):
    type: str = "Count"
    value: int = Field(None)


class QuantitySchema(AnyScalarComponentSchema):
    type: str = "Quantity"
    value: Union[float, str] = Field(None)
    uom: Union[UCUMCode, URI] = Field(...)

    @field_validator('value')
    @classmethod
    def validate_value(cls, v):
        if isinstance(v, Real):
            return v
        elif isinstance(v, str):
            if v in ['NaN', 'INFINITY', '+INFINITY', '-INFINITY']:
                return v
            else:
                raise ValueError(
                    'string representation of value must be one of the following: NaN, INFINITY, +INFINITY, -INFINITY')
        else:
            try:
                return float(v)
            except ValueError:
                raise ValueError('value must be a number or a string representing a special value '
                                 '[NaN, INFINITY, +INFINITY, -INFINITY]')


class TimeSchema(AnyScalarComponentSchema):
    type: str = "Time"
    value: str = Field(None)
    reference_time: str = Field(None, serialization_alias='referenceTime')
    local_frame: str = Field(None)
    uom: Union[UCUMCode, URI] = Field(...)


class CategorySchema(AnyScalarComponentSchema):
    type: str = "Category"
    value: str = Field(None)
    code_space: str = Field(None, serialization_alias='codeSpace')


class TextSchema(AnyScalarComponentSchema):
    type: str = "Text"
    value: str = Field(None)


class CountRangeSchema(AnySimpleComponentSchema):
    type: str = "CountRange"
    value: list[int] = Field(None)
    uom: Union[UCUMCode, URI] = Field(...)


class QuantityRangeSchema(AnySimpleComponentSchema):
    type: str = "QuantityRange"
    value: list[Union[float, str]] = Field(None)
    uom: Union[UCUMCode, URI] = Field(...)


class TimeRangeSchema(AnySimpleComponentSchema):
    type: str = "TimeRange"
    value: list[str] = Field(None)
    reference_time: str = Field(None, serialization_alias='referenceTime')
    local_frame: str = Field(None)
    uom: Union[UCUMCode, URI] = Field(...)


class CategoryRangeSchema(AnySimpleComponentSchema):
    type: str = "CategoryRange"
    value: list[str] = Field(None)
    code_space: str = Field(None, serialization_alias='codeSpace')



#  =============================================================================
#  Copyright (c) 2025 Botts Innovative Research Inc.
#  Date: 2025/10/1
#  Author: Ian Patterson
#  Contact Email: ian@botts-inc.com
#  =============================================================================

from pydantic import BaseModel, Field

from .csapi4py.constants import GeometryTypes


# TODO: Add specific validations for each type
# TODO: determine if serializing 'shapely' objects gives valid JSON structures from our own serialization
class Geometry(BaseModel):
    """
    A class to represent the geometry of a feature
    """
    type: GeometryTypes = Field(...)
    coordinates: list
    bbox: list = None



from src.oshconnect import OSHConnect, Node


def test_streamble_observations():
    app = OSHConnect("Test App")
    node = Node(address="localhost", port=8282, username="admin", password="admin", protocol="http")
    app.add_node(node)
    app.discover_systems()
    app.discover_datastreams()

    datastreams = app.get_datastreams()
    print(datastreams)


from pydantic import BaseModel, Field, ConfigDict


class Encoding(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    id: str = Field(None)
    type: str = Field(...)
    vector_as_arrays: bool = Field(False, alias='vectorAsArrays')


class JSONEncoding(Encoding):
    type: str = "JSONEncoding"



from datetime import datetime
from typing import Union, Optional, List

from pydantic import BaseModel, StrictStr, Field, field_validator


class QueryModel(BaseModel):
    id: list = None
    bbox: list = None
    date_time: Union[StrictStr, datetime] = Field(None, alias='datetime')
    geom: dict = None
    q: list = Optional[List[str]]
    parent: list = None
    procedure: list = None
    foi: list = None
    observed_property: list = Field(None, serialization_alias='observedProperty')
    controlled_property: list = Field(None, serialization_alias='controlledProperty')
    recursive: bool = False
    limit: int = Field(10, ge=1, le=10000)

    @field_validator('q')
    def validate_q(cls, v):
        if v is not None:
            return v.split(',')
        return v



from csapi4py.default_api_helpers import APIHelper

def test_url_generation():
    helper = APIHelper(server_url='localhost', port=8282, protocol='http', username='admin', password='admin', api_root='sensorhub/api')
    expected_url = "http://localhost:8282/sensorhub/api"
    url = helper.get_api_root_url()
    assert url == expected_url
    expected_url = "ws://localhost:8282/sensorhub/api"
    url = helper.get_api_root_url(socket=True)
    assert url == expected_url
    helper.set_protocol('https')
    expected_url = "https://localhost:8282/sensorhub/api"
    url = helper.get_api_root_url()
    assert url == expected_url
    expected_url = "wss://localhost:8282/sensorhub/api"
    url = helper.get_api_root_url(socket=True)
    assert url == expected_url



from pydantic import BaseModel, HttpUrl, Field


class TypeOf(BaseModel):
    """
    TypeOf is a resolvable reference to some other general process (that can be any type inheriting from AbstractProcess)
    :param href: The URL of the referenced process
    :param relationship: The relationship of the referenced process to the current process
    :param media_type: The media type of the referenced process
    :param href_lang: The language of the referenced process
    :param title: The title of the referenced process
    :param uid: The unique identifier of the referenced process
    :param target_resource: The target resource of the referenced process
    :param interface: The interface of the referenced process
    """
    href: HttpUrl
    relationship: str = Field(..., serialization_alias='rel')
    media_type: str = Field(None, serialization_alias='type')
    href_lang: str = Field(None, serialization_alias='hreflang')
    title: str = Field(None)
    uid: str = Field(None)
    target_resource: str = Field(None, serialization_alias='rt')
    interface: str = Field(None, serialization_alias='if')


class SMLAbstractProcess(BaseModel):
    description: str = None
    unique_id: str = Field(None, serialization_alias='uniqueID')
    label: str = None
    lang: str = None
    keywords: list = None
    identifiers: list = None
    classifiers: list = None
    valid_time: list = Field(None, serialization_alias='validTime')
    security_constraints: list = Field(None, serialization_alias='securityConstraints')
    legal_constraints: list = Field(None, serialization_alias='legalConstraints')
    characteristics: list = None
    capabilities: list = None
    contacts: list = None
    documents: list = None
    history: list = None
    definition: HttpUrl = None
    type_of: TypeOf = Field(None, serialization_alias='typeOf')
    configuration: HttpUrl = None
    features_of_interest: list = Field(None, serialization_alias='featuresOfInterest')
    inputs: list = None
    outputs: list = None
    parameters: list = None
    modes: list = None
    method: str = None
    position: list = None
